/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "bitvector.h"

#define WORDBITS (sizeof(unsigned) * 8)
#define NUM_WORDS(b) (((b) + (WORDBITS-1)) / WORDBITS)

#define TRUE 1
#define FALSE 0

/*-------------------------------------------------------------------------
  Allocate a new bit vector and set all of the bits to FALSE.  
-------------------------------------------------------------------------*/
bit_vector *new_bit_vector (int num_bits)
{
	int num_words = NUM_WORDS(num_bits);

	bit_vector *result = (bit_vector *)malloc(sizeof(bit_vector));

	assert(num_bits >= 0);

	result->num_bits = num_bits;
	result->bits = (unsigned *)calloc(num_words, sizeof(unsigned));

	return result;
}

/*-------------------------------------------------------------------------
 * 
-------------------------------------------------------------------------*/
void copy_n_bits_at_x_to_y(int nbits, bit_vector *from, int starting_at_bit_from, bit_vector *to, int starting_at_bit_to)
{
	int i;

	assert(from->num_bits >= starting_at_bit_from + nbits);
	assert(to->num_bits >= starting_at_bit_to + nbits);

	for (i = 0; i < nbits; i++)
	{
		set_bit(to, starting_at_bit_to+i, get_bit(from, starting_at_bit_from+i)); 
	}
}

/*-------------------------------------------------------------------------
 * 
-------------------------------------------------------------------------*/
void reverse_bits(bit_vector *vector)
{
	int i;
	bit_vector *temp = new_bit_vector(vector->num_bits);

	copy_bits (temp, vector);

	for (i = 0; i < vector->num_bits; i++)
	{
		set_bit(vector, i, get_bit(temp, vector->num_bits-1-i)); 
	}

	free_bit_vector(temp);
}

/*-------------------------------------------------------------------------
 * Change a set of bits in the vector to a short int value
-------------------------------------------------------------------------*/
void set_bits_with_short_with_mask_at(bit_vector *b, int ndx, unsigned short value, unsigned short inmask)
{
	int i;
	short mask = 1 << 7;
//	short mask = 1;

	for (i = 0; i < 8; i++)
	{
		if ((inmask & mask) != 0)
		{
			if((mask & value) != 0)
			{
				set_bit(b, ndx+i, TRUE);	
			}
			else
			{
				set_bit(b, ndx+i, FALSE);	
			}
		}
		mask = mask >> 1;
//		mask = mask << 1;
	}
}

/*-------------------------------------------------------------------------
 * Return an int from a bit vector
-------------------------------------------------------------------------*/
int get_int_from_bit_vector(bit_vector *b)
{
	int i;
	int return_int = 0;
//	int mask = 1 << b->num_bits;
	int mask = 1;

	assert (b->num_bits < 32);

	for (i = 0; i < b->num_bits; i++)
	{
		if (get_bit(b, i) == TRUE)
		{
			return_int += mask;
		}
		//mask = mask >> 1;
		mask = mask << 1;
	}	
	
	return return_int;
}

/*-------------------------------------------------------------------------
 * Change a set of bits in the vector to a int value for n bits
-------------------------------------------------------------------------*/
void set_n_bits_with_int(bit_vector *b, unsigned int value, int n)
{
	int i;
	short mask = 1;

	assert (n < 32);
	assert (b->num_bits >= n);

	for (i = 0; i < n; i++)
	{
		if((mask & value) != 0)
		{
			set_bit(b, i, TRUE);	
		}
		else
		{
			set_bit(b, i, FALSE);	
		}
		mask = mask << 1;
	}
}

/*-------------------------------------------------------------------------
 * Change a set of bits in the vector to a short int value
-------------------------------------------------------------------------*/
void set_bits_with_short_at(bit_vector *b, int ndx, unsigned short value)
{
	int i;
	short mask = 1 << 7;
//	short mask = 1;

	for (i = 0; i < 8; i++)
	{
		if((mask & value) != 0)
		{
			set_bit(b, ndx+i, TRUE);	
		}
		else
		{
			set_bit(b, ndx+i, FALSE);	
		}
		mask = mask >> 1;
//		mask = mask << 1;
	}
}

/*-------------------------------------------------------------------------
 * Get a set of bits into a unsigend short
-------------------------------------------------------------------------*/
unsigned short get_short_from_bits_at(bit_vector *b, int ndx, int number)
{
	int i;
	short return_short = 0;
	short mask = 1 << 7;
//	short mask = 1;

	if (number > 8)
	{
		return return_short;
	}

	for (i = 0; i < number; i++)
	{
		if (get_bit(b, ndx+i) == TRUE)
		{
			return_short += mask;
		}
		mask = mask >> 1;
//		mask = mask << 1;
	}	
	
	return return_short;
}

/*-------------------------------------------------------------------------
  Change the value of one bit in a vector.  The specified bit is set to
    the given digital_bit value.  
-------------------------------------------------------------------------*/
void set_bit (bit_vector *b, int ndx, digital_bit value)
{
	unsigned mask, old_bits;
	int word = ndx / WORDBITS;

	assert(ndx < b->num_bits);

	mask = 1 << (ndx % WORDBITS);
	old_bits = b->bits[word];

	if (value == FALSE) 
	{
		mask = ~mask;
		b->bits[word] = old_bits & mask;
	} 
	else 
	{
		b->bits[word] = old_bits | mask;
	}
}

/*-------------------------------------------------------------------------
  Retrieve the value of a bit in a vector.  
-------------------------------------------------------------------------*/
digital_bit get_bit (bit_vector *b, int ndx)
{
	unsigned mask;
	int word = ndx / WORDBITS;

	assert(ndx < b->num_bits);

	mask = 1 << (ndx % WORDBITS);
	return ((b->bits[word] & mask) != 0);
}

/*-------------------------------------------------------------------------
  Set all of the bits in a vector to one value.  
-------------------------------------------------------------------------*/
void set_all_bits (bit_vector *b, digital_bit value)
{
	int i;
	unsigned new_bits = 0;
	int num_words = NUM_WORDS(b->num_bits);

	if (value) 
		new_bits = ~new_bits;

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = new_bits;
	}
}

/*-------------------------------------------------------------------------
  Subtract bit vectors.  The result replaces the first vector (b = b - c).  
-------------------------------------------------------------------------*/
void subtract_bits (bit_vector *b, bit_vector *c)
{
	int i;
	unsigned old_bits, mask;
	int num_words = NUM_WORDS(b->num_bits);

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		old_bits = b->bits[i];
		mask = ~(c->bits[i]);
		b->bits[i] = old_bits & mask;
	}
}

/*-------------------------------------------------------------------------
  Boolean XOR of two bit vectors.  The result replaces the first vector.  
-------------------------------------------------------------------------*/
void xor_bits (bit_vector *b, bit_vector *c)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = (b->bits[i] ^ c->bits[i]);
	}
}

/*-------------------------------------------------------------------------
  Boolean AND of two bit vectors.  The result replaces the first vector.  
-------------------------------------------------------------------------*/
void and_bits (bit_vector *b, bit_vector *c)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = (b->bits[i] & c->bits[i]);
	}
}

/*-------------------------------------------------------------------------
  Boolean OR of two bit vectors.  The result replaces the first vector.  
-------------------------------------------------------------------------*/
void or_bits (bit_vector *b, bit_vector *c)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = (b->bits[i] | c->bits[i]);
	}
}

/*-------------------------------------------------------------------------
  Boolean NOT of bit vector.  The result replaces the first vector.  
-------------------------------------------------------------------------*/
void not_bits (bit_vector *b)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = ~b->bits[i];
	}
}

/*-------------------------------------------------------------------------
  Swap bits 0->1 to 1->0  
-------------------------------------------------------------------------*/
void swap_vectors (bit_vector *b, bit_vector *c)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;
	unsigned temp_bit;

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		temp_bit = b->bits[i];
		b->bits[i] = c->bits[i];
		c->bits[i] = temp_bit;
	}
}

/*-------------------------------------------------------------------------
  Copy bit vector "c" to bit vector "b".  
-------------------------------------------------------------------------*/
void copy_bits (bit_vector *b, bit_vector *c)
{
	int num_words = NUM_WORDS(b->num_bits);
	int i;

	assert(b->num_bits == c->num_bits);

	for (i = 0; i < num_words; i++) 
	{
		b->bits[i] = c->bits[i];
	}
}

/*-------------------------------------------------------------------------
  Test if two bit vectors are equal for n bits.  
-------------------------------------------------------------------------*/
digital_bit bits_are_equal_n_bits (bit_vector *b, bit_vector *c, int n)
{
	int n_full_words = n / WORDBITS;
	int n_excess_bits = n % WORDBITS;
	unsigned mask = (((unsigned)1) << n_excess_bits) - 1;
	int i;

	assert(b->num_bits >= n && c->num_bits >= n);

	/* check everything but the last word */
	for (i = 0; i < n_full_words; i++) 
	{
		if (b->bits[i] != c->bits[i]) 
			return FALSE;
	}

	/* check the remaining bits */
	if (n_excess_bits) 
	{
		if ((b->bits[n_full_words] & mask) != (c->bits[n_full_words] & mask)) 
		{
			return FALSE;
		}
	}

	return TRUE;
}
/*-------------------------------------------------------------------------
  Test if two bit vectors are equal.  
-------------------------------------------------------------------------*/
digital_bit bits_are_equal (bit_vector *b, bit_vector *c)
{
	int full_words = b->num_bits / WORDBITS;
	int excess_bits = b->num_bits % WORDBITS;
	unsigned mask = (((unsigned)1) << excess_bits) - 1;
	int i;

	assert(b->num_bits == c->num_bits);

	/* check everything but the last word */
	for (i = 0; i < full_words; i++) 
	{
		if (b->bits[i] != c->bits[i]) 
			return FALSE;
	}

	/* check the remaining bits */
	if (excess_bits) 
	{
		if ((b->bits[full_words] & mask) != (c->bits[full_words] & mask)) 
		{
			return FALSE;
		}
	}

	return TRUE;
}

/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
digital_bit bits_are_false (bit_vector *b)
{
	int full_words = b->num_bits / WORDBITS;
	int excess_bits = b->num_bits % WORDBITS;
	unsigned mask = (((unsigned)1) << excess_bits) - 1;
	int i;

	/* check everything but the last word */
	for (i = 0; i < full_words; i++) 
	{
		if (b->bits[i] != FALSE) 
			return FALSE;
	}

	/* check the remaining bits */
	if (excess_bits) 
	{
		if ((b->bits[full_words] & mask) != FALSE) 
		{
			return FALSE;
		}
	}

	return TRUE;
}

/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
digital_bit bits_are_true (bit_vector *b)
{
	int full_words = b->num_bits / WORDBITS;
	int excess_bits = b->num_bits % WORDBITS;
	unsigned mask = (((unsigned)1) << excess_bits) - 1;
	int i;

	/* check everything but the last word */
	for (i = 0; i < full_words; i++) 
	{
		if (b->bits[i] != TRUE)
			return FALSE;
	}

	/* check the remaining bits */
	if (excess_bits) 
	{
		if ((b->bits[full_words] & mask) != TRUE) 
		{
			return FALSE;
		}
	}

	return TRUE;
}

/*-------------------------------------------------------------------------
 count bits of given value 
-------------------------------------------------------------------------*/
unsigned count_bits (bit_vector *b, digital_bit v)
{
	int full_words = b->num_bits / WORDBITS;
	int excess_bits = b->num_bits % WORDBITS;
	unsigned mask = (((unsigned)1) << excess_bits) - 1;
	unsigned i,x;
	unsigned count = 0;

	for (i=0; i<full_words;i++) 
	{
		if (b->bits[i]) 
		{
			for (x=0;x < WORDBITS;x++)
			{
				if (((b->bits[i] & (1<<x)) >> x) == v)
					count++;
			}
		}
	}

	/* check remaining bits */
	if (excess_bits) 
	{
		if ((b->bits[full_words] & mask) != 0) 
		{
			for (x=0;x < excess_bits;x++)
			{
				if (((b->bits[i] & (1<<x)) >> x) == v)
					count++;
			}
		}
	}

	return count;
}

/*-------------------------------------------------------------------------
  Print the elements of a bit vector that are TRUE.  The bitprint argument
    specifies a function to print a single bit.  This allows you to print
    sensible output when the bits represent something other than integers.
    The fprint_bit function may be used to print them as integers.  
-------------------------------------------------------------------------*/
void fprint_bits (FILE *fd, bit_vector *b, bitprint_f bitprint)
{
	int i;

	for (i = 0; i < b->num_bits; i++) 
	{
		if (get_bit(b, i)) 
		{
			fputc(' ', fd);
			bitprint(fd, i);
		}
	}
}

/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void fprint_bits_out (FILE *fd, bit_vector *b, bitprint_f bitprint)
{
	int i;

	for (i = 0; i < b->num_bits; i++) 
	{
		printf("%d", get_bit(b, i));
	}
}

/*-------------------------------------------------------------------------
-------------------------------------------------------------------------*/
void fprint_bit (FILE *fd, int bit)
{
	fprintf(fd, "%d", bit);
}

/*-------------------------------------------------------------------------
  Deallocate a bit vector.  
-------------------------------------------------------------------------*/
void free_bit_vector (bit_vector *b)
{
	free(b->bits);
	free(b);
}
