/* 
Copyright (c) 2008 - Peter Jamieson

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_2dconv.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

conv_device_state_t conv_state;

void initialize_2dconv_state(conv_device_state_t *state);
void free_2dconv_state(conv_device_state_t *state);
void conv_process_event_clump(LINKED_LIST *output_timeline, conv_device_state_t *state, event_t *clump_to_process);

short Tdconv_do_next_convolution_step(conv_device_state_t *state, short old_is_output, int matrix_i, int matrix_j);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_2dconv)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_2dconv(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_2dconv_state(&conv_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		conv_process_event_clump(output_event_timeline, &conv_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_2dconv_state(&conv_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_2dconv_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_2dconv_state(conv_device_state_t *state)
{
	int i, j;

	state->is_convolution_matrix = FALSE;

#ifdef INT_CALC
	state->matrix = (int**)malloc(sizeof(int*)*MATRIX_SIZE);
	for (i = 0; i < MATRIX_SIZE; i++)
	{
		state->matrix[i] = (int*)malloc(sizeof(int)*MATRIX_SIZE);
	}

	state->out_matrix = (int**)malloc(sizeof(int*)*MATRIX_SIZE);
	for (i = 0; i < MATRIX_SIZE; i++)
	{
		state->out_matrix[i] = (int*)malloc(sizeof(int)*MATRIX_SIZE);
	}

	state->convolution_matrix = (int**)malloc(sizeof(int*)*CONVOLUTION_MATRIX_SIZE);
	for (i = 0; i < CONVOLUTION_MATRIX_SIZE; i++)
	{
		state->convolution_matrix[i] = (int*)malloc(sizeof(int)*CONVOLUTION_MATRIX_SIZE);
	}
#endif
#ifdef FIXEDPOINT_CALC
	state->matrix = (FIXED8_8**)malloc(sizeof(FIXED8_8*)*MATRIX_SIZE);
	for (i = 0; i < MATRIX_SIZE; i++)
	{
		state->matrix[i] = (FIXED8_8*)malloc(sizeof(FIXED8_8)*MATRIX_SIZE);
	}

	state->out_matrix = (FIXED8_8**)malloc(sizeof(FIXED8_8*)*MATRIX_SIZE);
	for (i = 0; i < MATRIX_SIZE; i++)
	{
		state->out_matrix[i] = (FIXED8_8*)malloc(sizeof(FIXED8_8)*MATRIX_SIZE);
	}

	state->convolution_matrix = (FIXED8_8**)malloc(sizeof(FIXED8_8*)*CONVOLUTION_MATRIX_SIZE);
	for (i = 0; i < CONVOLUTION_MATRIX_SIZE; i++)
	{
		state->convolution_matrix[i] = (FIXED8_8*)malloc(sizeof(FIXED8_8)*CONVOLUTION_MATRIX_SIZE);
	}
#endif
}

/*---------------------------------------------------------------------------------------------
 * (function: free_2dconv_state)
 *-------------------------------------------------------------------------------------------*/
void free_2dconv_state(conv_device_state_t *state)
{
}

/*---------------------------------------------------------------------------------------------
 * (function: conv_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void conv_process_event_clump(LINKED_LIST *output_timeline, conv_device_state_t *state, event_t *clump_to_process)
{
	int i, j, index_of_resource, k; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;
	short is_output = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* now go through all the events in the current clump which represents a timestep */
	for (index_of_resource = 0; index_of_resource < clump_to_process->num_resources; index_of_resource++)
	{	
		char *name = clump_to_process->resources[index_of_resource]->name;
		int num_values = clump_to_process->resources[index_of_resource]->num_values;
		char **value = clump_to_process->resources[index_of_resource]->value;
		short to_free;

		/* need to get theresource name for dfault names */
		if (clump_to_process->resources[index_of_resource]->name != NULL)
		{
			name = clump_to_process->resources[index_of_resource]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[index_of_resource]->default_name);
			to_free = TRUE;
		}

	
		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
			}
			else if (strcmp(value[0], "0") == 0)
			{
			}
		}
		else if (strcmp(name, "input_block") == 0)
		{
			/* IF - input data available then put it on the buffer */
			int current_index = 0;

			assert(strlen(value[0])/4 == CONVOLUTION_MATRIX_SIZE*CONVOLUTION_MATRIX_SIZE + MATRIX_SIZE*MATRIX_SIZE); // 32 bits

			for (i = 0; i < CONVOLUTION_MATRIX_SIZE; i++)
			{
				for (j = 0; j < CONVOLUTION_MATRIX_SIZE; j++)
				{
#ifdef INT_CALC
					state->convolution_matrix[i][j] = convert_hex_string_to_int(&(value[0][current_index]));
#endif
#ifdef FIXEDPOINT_CALC
					state->convolution_matrix[i][j].full = convert_hex_string_of_size_to_int(&(value[0][current_index]), 4);
#endif
					current_index += 4;
				}
			}

			for (i = 0; i < MATRIX_SIZE; i++)
			{
				for (j = 0; j < MATRIX_SIZE; j++)
				{
#ifdef INT_CALC
					state->matrix[i][j] = convert_hex_string_to_int(&(value[0][current_index]));
#endif
#ifdef FIXEDPOINT_CALC
					state->matrix[i][j].full = convert_hex_string_of_size_to_int(&(value[0][current_index]), 4);
#endif
					current_index += 4;
				}
			}

			/* start of the matrix */
			state->matrix_index = -1;
			state->matrix_process_index = 0;
		}
		else if (strcmp(name, "convolution_matrix") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
				state->is_convolution_matrix = TRUE;
			}
			else if (strcmp(value[0], "0") == 0)
			{
				state->is_convolution_matrix = FALSE;
			}
		}
		else if (strcmp(name, "input_data_valid") == 0)
		{
			/* If - the input data is valid and we are ready top process.  This is a simplified view. */
			char **output_values;
			int matrix_i = state->matrix_process_index / MATRIX_SIZE;
			int matrix_j = state->matrix_process_index % MATRIX_SIZE;
			
			/* process next piece of block */
			state->matrix_index++;

			/* can work on if 5*400 + 25 bits of data are in...actually could do a little earlier */
			if (state->matrix_index > (MATRIX_SIZE*CONVOLUTION_MATRIX_SIZE + CONVOLUTION_MATRIX_SIZE*CONVOLUTION_MATRIX_SIZE))
			{
				/* if we have buffered enough of the input matrix then start processing */
				is_output = Tdconv_do_next_convolution_step(state, FALSE, matrix_i, matrix_j);
			}
	
			if (is_output == TRUE)
			{
				/* put the data on the output */
				/* create the resource name */
				resource_name = (char*)malloc(sizeof(char)*10);
				sprintf(resource_name, "output");
		
				/* create the character array */
				output_values = (char**)malloc(sizeof(char*)*1);
				output_values[0] = (char*)malloc(sizeof(char)*9);
#ifdef INT_CALC
				sprintf(output_values[0], "%x", state->out_matrix[matrix_i][matrix_j]);
#endif
#ifdef FIXEDPOINT_CALC
				/* OVERFLOW just concatenates it so 16 bits are passed out */
				sprintf(output_values[0], "%x", OVERFLOW(state->out_matrix[matrix_i][matrix_j]));
#endif
		
				/* move the processed element index */
				state->matrix_process_index++;

				/* add this event to the output list since this results is an output event */
				add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
				output_time_used = TRUE;
			}
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: Tdconv_do_next_convolution_step)
 *-------------------------------------------------------------------------------------------*/
short Tdconv_do_next_convolution_step(conv_device_state_t *state, short old_is_output, int matrix_i, int matrix_j)
{
	int i, j;

	char debug_string[4096];
	short return_value = old_is_output;
	int max_neg = -(CONVOLUTION_MATRIX_SIZE/2); // -2 for default size 5...this is the adjustment so that 0,0 is -MATRIX_SIZE/2,-MATRIX_SIZE/2

	int next_i;
	int next_j;

	// Kernel flip calculations
	//
	// a b c d e   y x w v u
	// f g h i j   t s r q p
	// k l m n o   o n m l k
	// p q r s t   j i h g f
	// u v w x y   e d c b a
	//
	// 0,0 = m
	// -1,-1 = g
	// conversion on flip
	// 0,0 = 0,0
	// -2,-2 = 2,2
	// -1,1 = 1,-1
	// -2,0 = 2,0
	//
	// i,j = -i,-j 
	// but coordinbates are 0,0 = -2,-2 so flip is
	//
	// 0,0(a) => -2,-2 = 2,2 => 4,4 correct
	// 1,3(i) => -1,0 = 1,0 => 3,1 correct
	// i,j => i-2,j-2 = 2-i,2-j => 4-i,4-j
	//
	// Generalized = MATRIX_SIZE-index

#ifdef INT_CALC
	state->out_matrix[matrix_i][matrix_j] = 0;
#endif
#ifdef FIXEDPOINT_CALC
	state->out_matrix[matrix_i][matrix_j].full = 0;
#endif

	sprintf(debug_string, "-> ");
	
	for (i = 0; i < CONVOLUTION_MATRIX_SIZE; i++)	
	{
		for (j = 0; j < CONVOLUTION_MATRIX_SIZE; j++)
		{
			/* calculate the index in the matrix */
			next_i = matrix_i+(i+max_neg);
			next_j = matrix_j+(j+max_neg);

			if ((next_i < 0) || (next_i > MATRIX_SIZE - 1) || (next_j < 0) || (next_j > MATRIX_SIZE - 1))
			{
				/* IF - outside the boundaries of the matrix don't calculate */
				continue;
			}
			
#ifdef INT_CALC
			// do the calculations...Flip is MATRIX_SIZE-i and points are next_i (same for j)
			state->out_matrix[matrix_i][matrix_j] += state->convolution_matrix[CONVOLUTION_MATRIX_SIZE-1-i][CONVOLUTION_MATRIX_SIZE-1-j]*state->matrix[next_i][next_j]; 

			sprintf(debug_string, "%s+%d*%d", debug_string, state->convolution_matrix[CONVOLUTION_MATRIX_SIZE-1-i][CONVOLUTION_MATRIX_SIZE-1-j], state->matrix[next_i][next_j]);
#endif
#ifdef FIXEDPOINT_CALC
// do the calculations...Flip is MATRIX_SIZE-i and points are next_i (same for j)
			state->out_matrix[matrix_i][matrix_j].full += MULT8_8(state->convolution_matrix[CONVOLUTION_MATRIX_SIZE-1-i][CONVOLUTION_MATRIX_SIZE-1-j],state->matrix[next_i][next_j]); 

			sprintf(debug_string, "%s+%x.%x*%x.%x", debug_string, state->convolution_matrix[CONVOLUTION_MATRIX_SIZE-1-i][CONVOLUTION_MATRIX_SIZE-1-j].part.integer, state->convolution_matrix[CONVOLUTION_MATRIX_SIZE-1-i][CONVOLUTION_MATRIX_SIZE-1-j].part.fraction, state->matrix[next_i][next_j].part.integer, state->matrix[next_i][next_j].part.fraction);
#endif
		}
	}

	if ((matrix_i == 0) && (matrix_j == 0)) // set to some matrix value to check if correct convolution
	{
		printf("%s\n", debug_string);
	}

	/* we've generated an output */
	return_value = TRUE;

	return return_value;
}
