/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_aes.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

ae_device_state_t ae_state;

void initialize_aes_state(ae_device_state_t *state);
void free_aes_state(ae_device_state_t *state);
void ae_process_event_clump(LINKED_LIST *output_timeline, ae_device_state_t *state, event_t *clump_to_process);

// Keyexpansion routines
bit_vector **do_key_expansion(int Nb, int Nr, int Nk, ae_device_state_t *state);
void subBytes(bit_vector *vector);
void subWord(bit_vector *vector);
void rotWord(bit_vector *vector);
bit_vector *Rcon(int i);

// encryption routines
void aes_encrypt_next_block(ae_device_state_t *state);
void AddRoundKey(int Nb, int Nr, bit_vector ***state_array, bit_vector **key_schedule);
void SubBytes(int Nb, bit_vector ***state_array);
void ShiftRows(int Nb, bit_vector ***state_array);
void MixColumns(int Nb, bit_vector ***state_array);
void gmix_column(unsigned char *r);

// decryption routines
void aes_decrypt_next_block(ae_device_state_t *state);
void InvMixColumns(int Nb, bit_vector ***state_array);
void ginvmix_column(unsigned char *r);
void InvShiftRows(int Nb, bit_vector ***state_array);
void InvSubBytes(int Nb, bit_vector ***state_array);

void debug_state_array(int Nb, bit_vector***state_array);
void debug_keyschedule(int Nb, int Nr, bit_vector **key_schedule);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_aes)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_aes(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_aes_state(&ae_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		ae_process_event_clump(output_event_timeline, &ae_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_aes_state(&ae_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_aes_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_aes_state(ae_device_state_t *state)
{
	state->key = new_bit_vector(128);
	state->input_data = new_bit_vector(128);
	state->input_block_data = NULL;
	state->num_input_block_data = 0;
	state->block_size = 0;
	state->is_input_block = -99;

	state->output_data = new_bit_vector(128);
}

/*---------------------------------------------------------------------------------------------
 * (function: free_aes_state)
 *-------------------------------------------------------------------------------------------*/
void free_aes_state(ae_device_state_t *state)
{
	free_bit_vector(state->key);
	free_bit_vector(state->input_data);
	free_bit_vector(state->output_data);
}

/*---------------------------------------------------------------------------------------------
 * (function: ae_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void ae_process_event_clump(LINKED_LIST *output_timeline, ae_device_state_t *state, event_t *clump_to_process)
{
	int i, index_of_resource, k; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* now go through all the events in the current clump which represents a timestep */
	for (index_of_resource = 0; index_of_resource < clump_to_process->num_resources; index_of_resource++)
	{	
		char *name = clump_to_process->resources[index_of_resource]->name;
		int num_values = clump_to_process->resources[index_of_resource]->num_values;
		char **value = clump_to_process->resources[index_of_resource]->value;
		short to_free;

		/* need to get theresource name for dfault names */
		if (clump_to_process->resources[index_of_resource]->name != NULL)
		{
			name = clump_to_process->resources[index_of_resource]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[index_of_resource]->default_name);
			to_free = TRUE;
		}


		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
				/* set the key to 0 */
				set_all_bits(state->key, FALSE);
				state->is_input_block = -99;
			}
			else if (strcmp(value[0], "0") == 0)
			{
			}
		}
		else if (strcmp(name, "key") == 0)
		{
			/* IF - key is there then store it */
			short next_short;

			for (i = 0; i < 16; i++)
			{
				next_short = convert_hex_string_to_short(&(value[0][i*2]), 2);
				set_bits_with_short_at(state->key, i*8, next_short);
			}
		}
		else if (strcmp(name, "input_block") == 0)
		{
			/* IF - input data available then put it on the buffer */
			short next_short;

			assert(strlen(value[0]) % 32 == 0);
			assert(state->num_input_block_data == 0);

			state->num_input_block_data = strlen(value[0]) / 32;
			state->top_input_block_data = 0;
			state->input_block_data = (bit_vector **)malloc(sizeof(bit_vector*)*state->num_input_block_data);

			/* read every two chars from the input and convert to bit vector form */
			for (i = 0; i < state->num_input_block_data; i++)
			{
				state->input_block_data[i] = new_bit_vector(128);

				for (k = 0; k < 16; k++)
				{
					next_short = convert_hex_string_to_short(&(value[0][i*32+k*2]), 2);
					set_bits_with_short_at(state->input_block_data[i], k*8, next_short);
				}
			}

			assert(state->is_input_block == -99);
			state->is_input_block = TRUE;
		}
		else if (strcmp(name, "input_data_valid") == 0)
		{
			/* If - the input data is valid and we are ready top process.  This is a simplified view. */
			char **output_values;

			assert(state->is_input_block == TRUE || state->is_input_block == FALSE);
			
			state->block_size += 4;

			/* need 128 bits which is = to 8 data valid signals */
			if (state->block_size == 32)
			{
				/* when we've read all 32 hex values then we can process a 128 bit block */
				state->block_size = 0;

				if (state->is_input_block == TRUE)
				{
					/* process next piece of block */
					copy_n_bits_at_x_to_y(128, state->input_block_data[state->top_input_block_data], 0, state->input_data, 0);
					state->top_input_block_data ++;
	
					if (state->top_input_block_data == state->num_input_block_data)
					{
						/* IF - the block is completely processed free and reset for next block */
						for (i = 0; i < state->num_input_block_data; i++)
						{
							free_bit_vector(state->input_block_data[i]);
						}
						free(state->input_block_data);
						state->input_block_data = NULL;
						state->num_input_block_data = 0;
	
						state->is_input_block = -99;
					}
				}
	
				/* do the encryption and a check */
				aes_encrypt_next_block(state);
				aes_decrypt_next_block(state); // we do decryption to verify that the encryption algorithm is working
		
				/* put the data on the output */
		
				/* create the resource name */
				resource_name = (char*)malloc(sizeof(char)*10);
				sprintf(resource_name, "output");
		
				/* create the character array */
				output_values = (char**)malloc(sizeof(char*)*1);
				output_values[0] = convert_bitvector_to_hex_string(state->output_data);
		
				/* add this event to the output list since this results is an output event */
				add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
				output_time_used = TRUE;
			}
		}
		else
		{
			/* invalid input */
			assert(FALSE);
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: aes_encrypt_next_block)
 *  Nb = 4, Nk = 4, Nr = 10
 *  To understand the algorithm you'll need to look at FIPS197
 *-------------------------------------------------------------------------------------------*/
void aes_encrypt_next_block(ae_device_state_t *state)
{
	int i, j;
	int Nb = 4;
	int Nk = 4;
	int Nr = 10;
	bit_vector ***state_array;

	/* first do the the Key expansion as of pg19 FIPS-197 */
	bit_vector **key_schedule = do_key_expansion(Nb, Nr, Nk, &ae_state);
	//debug_keyschedule(Nb, Nr, key_schedule);

	/* copy in the input into the state based on FIPS0197 fromat in 3.4 ... note that i represents rows and j is columns.  This
	 * means that the first indexing element is row based and I will maintain this standard */
	state_array = (bit_vector***)malloc(sizeof(bit_vector**)*4);
	for (i = 0; i < 4; i++)
	{
		state_array[i] = (bit_vector**)malloc(sizeof(bit_vector*)*Nb);

		for (j = 0; j < Nb; j++)
		{
			state_array[i][j]= new_bit_vector(8);

			/* copying in the bits where i is row and j is collumn */
			copy_n_bits_at_x_to_y(8, state->input_data, i*32+j*8, state_array[i][j], 0); 
		}
			
	}

	/* do the roundkey */
	AddRoundKey(Nb, 0, state_array, key_schedule);

	/* do the rounds of AES */
	for (i = 1; i < Nr; i++)
	{
		SubBytes(Nb, state_array);		
		ShiftRows(Nb, state_array);
		MixColumns(Nb, state_array);
		AddRoundKey(Nb, i, state_array, key_schedule);
	}

	SubBytes(Nb, state_array);		
	ShiftRows(Nb, state_array);
	AddRoundKey(Nb, Nr, state_array, key_schedule);

	/* copy out the output */
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			/* copying in the bits where i is row and j is collumn */
			copy_n_bits_at_x_to_y(8, state_array[i][j], 0, state->output_data, i*32+j*8); 
		}
	}

	/* clean up */
	for (i = 0; i < (Nb*(Nr+1)); i++)
	{
		free_bit_vector(key_schedule[i]);
	}
	free(key_schedule);
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			free(state_array[i][j]);
		}
		free(state_array[i]);
	}
	free(state_array);
}

/*---------------------------------------------------------------------------------------------
 * (function: MixColumns)
 *-------------------------------------------------------------------------------------------*/
void MixColumns(int Nb, bit_vector ***state_array)
{
	int i, j;
	char bit_array[4];

	/* need to put the bitvectors into a char array for the gmix_collumn function
	 * that performs the Galoa field multiplication */
	for (i = 0; i < Nb; i++)
	{
		for (j = 0; j < 4; j++)
		{
			bit_array[j] = (char)get_short_from_bits_at(state_array[i][j], 0, 8);
		}

		gmix_column(bit_array); 

		for (j = 0; j < 4; j++)
		{
			set_bits_with_short_at(state_array[i][j], 0, (short)bit_array[j]);
		}
	}	
}

/*---------------------------------------------------------------------------------------------
 * (function: gmix_column)
 * Taken from wikipedia and slightly modified
 *-------------------------------------------------------------------------------------------*/
void gmix_column(unsigned char *r) 
{
	unsigned char a[4];
	unsigned char b[4];
	unsigned char c;
	unsigned char h;

        /* The array 'a' is simply a copy of the input array 'r'
	* The array 'b' is each element of the array 'a' multiplied by 2
	* in Rijndael's Galois field
	* a[n] ^ b[n] is element n multiplied by 3 in Rijndael's Galois field */ 
	for(c=0;c<4;c++) 
	{
		a[c] = r[c];
		h = r[c] & 0x80; /* hi bit */
		b[c] = r[c] << 1;

		if(h == 0x80) 
		{
			b[c] ^= 0x1b; /* Rijndael's Galois field */
		}
	}

	r[0] = b[0] ^ a[3] ^ a[2] ^ b[1] ^ a[1]; /* 2 * a0 + a3 + a2 + 3 * a1 */
	r[1] = b[1] ^ a[0] ^ a[3] ^ b[2] ^ a[2]; /* 2 * a1 + a0 + a3 + 3 * a2 */
	r[2] = b[2] ^ a[1] ^ a[0] ^ b[3] ^ a[3]; /* 2 * a2 + a1 + a0 + 3 * a3 */
	r[3] = b[3] ^ a[2] ^ a[1] ^ b[0] ^ a[0]; /* 2 * a3 + a2 + a1 + 3 * a0 */
}

/*---------------------------------------------------------------------------------------------
 * (function: ShiftRows)
 *-------------------------------------------------------------------------------------------*/
void ShiftRows(int Nb, bit_vector ***state_array)
{
	int i,j;
	bit_vector *temp_vector = new_bit_vector(Nb*8);	

	/* for 4 rows we're doing a shift by ith row to the left */
	for (i = 1; i < 4; i++)
	{
		/* copy the bits into a temp vector */
		for (j = 0; j < Nb; j++)
		{
			copy_n_bits_at_x_to_y(8, state_array[j][i], 0, temp_vector, j*8); 
		}
		
		/* shift by i spots to the left...0 becomes Nb - 1...1 becomes 0... and so on if i = 1 */
		/* copy back into the right spot */
		for (j = 0; j < Nb; j++)
		{
			copy_n_bits_at_x_to_y(8, temp_vector, j*8, state_array[(Nb+j-i)%Nb][i], 0); 
		}
	}

	free_bit_vector (temp_vector);
}

/*---------------------------------------------------------------------------------------------
 * (function: SubBytes)
 *-------------------------------------------------------------------------------------------*/
void SubBytes(int Nb, bit_vector ***state_array)
{
	int i, j;

	/* for each element in the state do a substitution based on the SBox */
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			short byte = get_short_from_bits_at(state_array[i][j], 0, 8);
			short sbox_val = getSBoxValue(byte);
			set_bits_with_short_at(state_array[i][j], 0, sbox_val);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: AddRoundKey)
 *-------------------------------------------------------------------------------------------*/
void AddRoundKey(int Nb, int round, bit_vector ***state_array, bit_vector **key_schedule)
{
	int i;
	bit_vector *current_key = new_bit_vector(Nb*8);

	/* for each byte do an xor with the key_schedule */
	for (i = 0; i < Nb; i++)
	{
		copy_n_bits_at_x_to_y(8, state_array[i][0], 0, current_key, 0); 
		copy_n_bits_at_x_to_y(8, state_array[i][1], 0, current_key, 8); 
		copy_n_bits_at_x_to_y(8, state_array[i][2], 0, current_key, 16); 
		copy_n_bits_at_x_to_y(8, state_array[i][3], 0, current_key, 24); 

		xor_bits(current_key, key_schedule[round*Nb+i]);

		copy_n_bits_at_x_to_y(8, current_key, 0, state_array[i][0], 0); 
		copy_n_bits_at_x_to_y(8, current_key, 8, state_array[i][1], 0); 
		copy_n_bits_at_x_to_y(8, current_key, 16, state_array[i][2], 0); 
		copy_n_bits_at_x_to_y(8, current_key, 24, state_array[i][3], 0); 
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: do_key_expansion)
 * K = 128, 4*11 (44) words generated
 *-------------------------------------------------------------------------------------------*/
bit_vector **do_key_expansion(int Nb, int Nr, int Nk, ae_device_state_t *state)
{
	int index = 0;
	int i;
	bit_vector **key_schedule;	
	bit_vector *temp_vector = new_bit_vector(32);	
	bit_vector *rotVector;

	key_schedule = (bit_vector**)malloc(sizeof(bit_vector*)*(Nb*(Nr+1))); // Nb (Nr + 1) words in array
	for (i = 0; i < (Nb*(Nr+1)); i++)
	{
		key_schedule[i] = new_bit_vector(32); // 4-byte words
	}

	while(index < Nk)
	{
		/* key_schedule = {key[4*i], key[4*i+1], key[4*i+2], key[4*i+3]} */
		copy_n_bits_at_x_to_y(32, state->key, 32*index, key_schedule[index], 0);
		index++;
	}

	index = Nk;

	while (index < (Nb*(Nr+1)))
	{
		copy_n_bits_at_x_to_y(32, key_schedule[index-1], 0, temp_vector, 0); 

		if (index % Nk == 0)
		{
			/* SubWord(RotWord(temp)) xor Rcon[i/Nk] */
			rotWord(temp_vector);
			subWord(temp_vector);			
			rotVector = Rcon(index/Nk);
			xor_bits(temp_vector, rotVector);

			free_bit_vector(rotVector);
		}
		else if ((Nk > 6) && (index % Nk == 4))
		{
			/* doesn't apply for our 128 bit, but I'm adding anyway */	
			subWord(temp_vector);			
		}

		/* w[i] = w[i-Nk] xor temp */
		xor_bits(temp_vector, key_schedule[index-Nk]);
		copy_n_bits_at_x_to_y(32, temp_vector, 0, key_schedule[index], 0); 

		index++;
	}		

	/* free the temp */
	free_bit_vector (temp_vector);

	/* return key schedule */
	return key_schedule;
}

/*---------------------------------------------------------------------------------------------
 * (function: Rcon)
 *-------------------------------------------------------------------------------------------*/
bit_vector *Rcon(int i)
{
	bit_vector *return_vector = new_bit_vector(32);	

	/* set all bits to 0 */
	set_all_bits (return_vector, FALSE);

	/* set the appropriate bit...this is a simple subsitution technique */
	if (i == 1)
	{
		/* 01000000 */ 
		set_bit(return_vector, 7, TRUE);
	}
	else if (i == 2)
	{
		/* 02000000 */ 
		set_bit(return_vector, 6, TRUE);
	}
	else if (i == 3)
	{
		/* 04000000 */ 
		set_bit(return_vector, 5, TRUE);
	}
	else if (i == 4)
	{
		/* 08000000 */ 
		set_bit(return_vector, 4, TRUE);
	}
	else if (i == 5)
	{
		/* 10000000 */ 
		set_bit(return_vector, 3, TRUE);
	}
	else if (i == 6)
	{
		/* 20000000 */ 
		set_bit(return_vector, 2, TRUE);
	}
	else if (i == 7)
	{
		/* 40000000 */ 
		set_bit(return_vector, 1, TRUE);
	}
	else if (i == 8)
	{
		/* 80000000 */ 
		set_bit(return_vector, 0, TRUE);
	}
	else if (i == 9)
	{
		/* 1b000000 */ 
		set_bit(return_vector, 3, TRUE);
		set_bit(return_vector, 4, TRUE);
		set_bit(return_vector, 6, TRUE);
		set_bit(return_vector, 7, TRUE);
	}
	else if (i == 10)
	{
		/* 36000000 */ 
		set_bit(return_vector, 2, TRUE);
		set_bit(return_vector, 3, TRUE);
		set_bit(return_vector, 5, TRUE);
		set_bit(return_vector, 6, TRUE);
	}

	return return_vector;
}

/*---------------------------------------------------------------------------------------------
 * (function: rotWord)
 *-------------------------------------------------------------------------------------------*/
void rotWord(bit_vector *vector)
{
	bit_vector *temp_vector = new_bit_vector(32);	

	/* do a set rotation by copying into to temp and then back out */
	copy_n_bits_at_x_to_y(32, vector, 0, temp_vector, 0); 

	copy_n_bits_at_x_to_y(8, temp_vector, 0, vector, 24);
	copy_n_bits_at_x_to_y(8, temp_vector, 8, vector, 0);
	copy_n_bits_at_x_to_y(8, temp_vector, 16, vector, 8);
	copy_n_bits_at_x_to_y(8, temp_vector, 24, vector, 16);

	free_bit_vector (temp_vector);
}

/*---------------------------------------------------------------------------------------------
 * (function: subWord)
 *-------------------------------------------------------------------------------------------*/
void subWord(bit_vector *vector)
{
	int i;

	/* substitute words based on the SBOX */
	for (i = 0; i < 4; i++)
	{
		short byte = get_short_from_bits_at(vector, i*8, 8);
		short sbox_val = getSBoxValue(byte);
		set_bits_with_short_at(vector, i*8, sbox_val);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: getSBoxValue)
 * Taken from Wikipedia
 *-------------------------------------------------------------------------------------------*/
int getSBoxValue(int num)
{
	static const int sbox[256] =   {
		//0     1     2     3     4     5     6     7     8     9     A     B     C     D     E     F
		0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76, //0
		0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0, //1
		0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15, //2
		0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75, //3
		0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84, //4
		0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf, //5
		0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8, //6
		0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2, //7
		0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73, //8
		0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb, //9
		0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79, //A
		0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08, //B
		0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a, //C
		0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e, //D
		0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf, //E
		0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16 }; //F
	return sbox[num];
}


/*---------------------------------------------------------------------------------------------
 * (function: aes_decrypt_next_block)
 *  Nb = 4, Nk = 4, Nr = 10
 *  To understand the algorithm you'll need to look at FIPS197
 *-------------------------------------------------------------------------------------------*/
void aes_decrypt_next_block(ae_device_state_t *state)
{
	int i, j;
	int Nb = 4;
	int Nk = 4;
	int Nr = 10;
	bit_vector ***state_array;
	bit_vector *temp_vector = new_bit_vector(Nb*4*8);	
	char *temp_string;

	/* first do the the Key expansion as of pg19 FIPS-197 */
	bit_vector **key_schedule = do_key_expansion(Nb, Nr, Nk, &ae_state);

	/* copy in the input into the state.  Same format as encryption. */
	state_array = (bit_vector***)malloc(sizeof(bit_vector**)*4);
	for (i = 0; i < 4; i++)
	{
		state_array[i] = (bit_vector**)malloc(sizeof(bit_vector*)*Nb);

		for (j = 0; j < Nb; j++)
		{
			state_array[i][j]= new_bit_vector(8);

			/* copying in the bits where i is row and j is collumn */
			copy_n_bits_at_x_to_y(8, state->output_data, i*32+j*8, state_array[i][j], 0); 
		}
			
	}

	/* do the roundkey */
	AddRoundKey(Nb, Nr, state_array, key_schedule);

	for (i = Nr-1; i > 0; i--)
	{
		InvShiftRows(Nb, state_array);
		InvSubBytes(Nb, state_array);		
		AddRoundKey(Nb, i, state_array, key_schedule);
		InvMixColumns(Nb, state_array);
	}

	InvShiftRows(Nb, state_array);
	InvSubBytes(Nb, state_array);		
	AddRoundKey(Nb, 0, state_array, key_schedule);

	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			/* copying in the bits where i is row and j is collumn */
			copy_n_bits_at_x_to_y(8, state_array[i][j], 0, temp_vector, i*32+j*8); 
		}
	}
	/* show the final string.  It should match the input string */
	printf("decrypted string %s\n", temp_string = convert_bitvector_to_hex_string(temp_vector));

	/* clean up */
	free(temp_string);
	free_bit_vector(temp_vector);
	for (i = 0; i < (Nb*(Nr+1)); i++)
	{
		free_bit_vector(key_schedule[i]);
	}
	free(key_schedule);
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			free(state_array[i][j]);
		}
		free(state_array[i]);
	}
	free(state_array);

}
		
/*---------------------------------------------------------------------------------------------
 * (function: InvMixColumns)
 *-------------------------------------------------------------------------------------------*/
void InvMixColumns(int Nb, bit_vector ***state_array)
{
	int i, j;
	char bit_array[4];

	/* copy into a string format for the ginvmix_collumn that does the InvMixColumn multiplcations */
	for (i = 0; i < Nb; i++)
	{
		for (j = 0; j < 4; j++)
		{
			bit_array[j] = (char)get_short_from_bits_at(state_array[i][j], 0, 8);
		}

		ginvmix_column(bit_array); 

		for (j = 0; j < 4; j++)
		{
			set_bits_with_short_at(state_array[i][j], 0, (short)bit_array[j]);
		}
	}	
}

// xtime is a macro that finds the product of {02} and the argument to xtime modulo {1b} 
#define xtime(x)   ((x<<1) ^ (((x>>7) & 1) * 0x1b))

// Multiplty is a macro used to multiply numbers in the field GF(2^8)
#define Multiply(x,y) (((y & 1) * x) ^ ((y>>1 & 1) * xtime(x)) ^ ((y>>2 & 1) * xtime(xtime(x))) ^ ((y>>3 & 1) * xtime(xtime(xtime(x)))) ^ ((y>>4 & 1) * xtime(xtime(xtime(xtime(x))))))

// MixColumns function mixes the columns of the state matrix.
// The method used to multiply may be difficult to understand for beginners.
// Please use the references to gain more information.
// PAJ - borrowed from http://www.hoozi.com/Articles/AES_Decryption.htm
void ginvmix_column(unsigned char *r) 
{
	int i;
	unsigned char a,b,c,d;
	a = r[0];
	b = r[1];
	c = r[2];
	d = r[3];
	r[0] = Multiply(a, 0x0e) ^ Multiply(b, 0x0b) ^ Multiply(c, 0x0d) ^ Multiply(d, 0x09);
	r[1] = Multiply(a, 0x09) ^ Multiply(b, 0x0e) ^ Multiply(c, 0x0b) ^ Multiply(d, 0x0d);
	r[2] = Multiply(a, 0x0d) ^ Multiply(b, 0x09) ^ Multiply(c, 0x0e) ^ Multiply(d, 0x0b);
	r[3] = Multiply(a, 0x0b) ^ Multiply(b, 0x0d) ^ Multiply(c, 0x09) ^ Multiply(d, 0x0e);
}

/*---------------------------------------------------------------------------------------------
 * (function: InvShiftRows)
 *-------------------------------------------------------------------------------------------*/
void InvShiftRows(int Nb, bit_vector ***state_array)
{
	int i,j;
	bit_vector *temp_vector = new_bit_vector(Nb*8);	

	/* does a reverse rotate right by i spots */
	for (i = 1; i < 4; i++)
	{
		/* copy the bits into a temp vector */
		for (j = 0; j < Nb; j++)
		{
			copy_n_bits_at_x_to_y(8, state_array[j][i], 0, temp_vector, j*8); 
		}
		
		/* shift by i spots to the left...0 becomes Nb - 1...1 becomes 0... and so on if i = 1 */
		/* copy back into the right spot */
		for (j = 0; j < Nb; j++)
		{
			copy_n_bits_at_x_to_y(8, temp_vector, j*8, state_array[(j+i)%Nb][i], 0); 
		}
	}

	free_bit_vector (temp_vector);
}

/*---------------------------------------------------------------------------------------------
 * (function: InvSubBytes)
 *-------------------------------------------------------------------------------------------*/
void InvSubBytes(int Nb, bit_vector ***state_array)
{
	int i, j;

	/* use the inverse SBOX inversion to convert back to the original version */
	for (i = 0; i < 4; i++)
	{
		for (j = 0; j < Nb; j++)
		{
			short byte = get_short_from_bits_at(state_array[i][j], 0, 8);
			short sbox_val = getSBoxInvert(byte);
			set_bits_with_short_at(state_array[i][j], 0, sbox_val);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: getSBoxInvert)
 * Taken from Wikipedia
 *-------------------------------------------------------------------------------------------*/
int getSBoxInvert(int num)
{
	static const int rsbox[256] = {
		0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb,
		0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb,
		0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
		0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25,
		0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92,
	        0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
		0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06,
		0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b,
		0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
		0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e,
		0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b,
		0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
		0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f,
		0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef,
		0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
		0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d
	    };
	   
    return rsbox[num];
}

void debug_keyschedule(int Nb, int Nr, bit_vector **key_schedule)
{
	int i, j;
	char *debug;

	for (i = 0; i < Nb*(Nr+1); i++)
	{
		printf("%s\n", debug = convert_bitvector_to_hex_string(key_schedule[i]));
		free(debug);
	}
}

void debug_state_array(int Nb, bit_vector***state_array)
{
	int i, j;
	char *debug;

	for (j = 0; j < Nb; j++)
	{
		for (i = 0; i < 4; i++)
		{
			printf("%s ", debug = convert_bitvector_to_hex_string(state_array[i][j]));
			free(debug);
		}
		printf("\n");
	}
	printf("\n");
}
