/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_bridge.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

conv_device_state_t conv_state;

void initialize_bridge_state(conv_device_state_t *state);
void free_bridge_state(conv_device_state_t *state);
void bridge_process_event_clump(LINKED_LIST *output_timeline, conv_device_state_t *state, event_t *clump_to_process);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_bridge)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_bridge(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_bridge_state(&conv_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		bridge_process_event_clump(output_event_timeline, &conv_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_bridge_state(&conv_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_bridge_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_bridge_state(conv_device_state_t *state)
{
}

/*---------------------------------------------------------------------------------------------
 * (function: free_bridge_state)
 *-------------------------------------------------------------------------------------------*/
void free_bridge_state(conv_device_state_t *state)
{
}

/*---------------------------------------------------------------------------------------------
 * (function: bridge_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void bridge_process_event_clump(LINKED_LIST *output_timeline, conv_device_state_t *state, event_t *clump_to_process)
{
	int i, j, index_of_resource, k; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;
	short is_output = FALSE;

	/* setup the output time delay */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* now go through all the events in the current clump which represents a timestep */
	for (index_of_resource = 0; index_of_resource < clump_to_process->num_resources; index_of_resource++)
	{	
		char *name = clump_to_process->resources[index_of_resource]->name;
		int num_values = clump_to_process->resources[index_of_resource]->num_values;
		char **value = clump_to_process->resources[index_of_resource]->value;
		short to_free;

		/* need to get the resource name for default names */
		if (clump_to_process->resources[index_of_resource]->name != NULL)
		{
			name = clump_to_process->resources[index_of_resource]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[index_of_resource]->default_name);
			to_free = TRUE;
		}
	
		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
			}
			else if (strcmp(value[0], "0") == 0)
			{
			}
		}
		else if (strcmp(name, "read") == 0)
		{
			/* This is a read statement from the parallel bus */
			char **spi_out;
			int spi_bit_stream = atoi(value[0]); // MSb is 0 for read, and the next 7 bits are value
			mpz_t *time_to_add;

			/* send out the request to the SPI slave */
			spi_out = (char**)malloc(sizeof(char*)*1);
			spi_out[0] = convert_int_to_bit_string(spi_bit_stream, 8);

			/* create the resource name */
			resource_name = (char*)malloc(sizeof(char)*7);
			sprintf(resource_name, "spi_SO");

			/* add this event to the output list since this results is an output event */
			add_event_to_timeline(output_timeline, output_time, resource_name, spi_out, 1, OUTPUT_EVENT); 
			output_time_used = TRUE;

			/* send back the read information the parallel master */
			time_to_add = (mpz_t*)malloc(sizeof(mpz_t));
			mpz_init(*time_to_add);
			/* adds some time for the reading to happen */
			mpz_add_ui(*time_to_add, *output_time, (unsigned int)(1000));

			/* send out the request to the SPI slave */
			spi_out = (char**)malloc(sizeof(char*)*1);
			spi_out[0] = (char*)malloc(sizeof(char)*7);
			sprintf(spi_out[0], "%d", state->address_memory[spi_bit_stream]);

			resource_name = (char*)malloc(sizeof(char)*10);
			sprintf(resource_name, "DAT_O");

			add_event_to_timeline(output_timeline, time_to_add, resource_name, spi_out, 1, OUTPUT_EVENT); 
		}
		else if (strcmp(name, "write") == 0)
		{
			/* This is a write statement from the parallel bus */
			char **spi_out;
			int spi_bit_stream = (1 << 23) | (atoi(value[0])<<16) | (atoi(value[1])); // MSb is 1 for write, and the next 7 bits are address, and the last 16 bits are the data
			mpz_t *time_to_add;

			/* send out the request to the SPI slave */
			spi_out = (char**)malloc(sizeof(char*)*1);
			spi_out[0] = convert_int_to_bit_string(spi_bit_stream, 24);

			/* create the resource name */
			resource_name = (char*)malloc(sizeof(char)*7);
			sprintf(resource_name, "spi_SO");

			/* add this event to the output list since this results is an output event */
			add_event_to_timeline(output_timeline, output_time, resource_name, spi_out, 1, OUTPUT_EVENT); 
			output_time_used = TRUE;
			
			/* record the value */
			assert(atoi(value[0]) < BRIDGE_ADDRESS_RANGE);
			state->address_memory[atoi(value[0])] = atoi(value[1]);
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}
