/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_glue_logic.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

gl_device_state_t glue_state;

void initialize_glue_state(gl_device_state_t *state);
void free_glue_state(gl_device_state_t *state);
void gl_process_event_clump(LINKED_LIST *output_timeline, gl_device_state_t *state, event_t *clump_to_process);
short send_outputs_depending_on_state(mpz_t *output_time, LINKED_LIST *output_timeline, short output_time_used, short state, int counter);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_glue_logic)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_glue_logic(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_glue_state(&glue_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		gl_process_event_clump(output_event_timeline, &glue_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_glue_state(&glue_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_glue_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_glue_state(gl_device_state_t *state)
{
	/* initialize all the output registers */
	state->counter1 = new_bit_vector(12);
	state->counter2 = new_bit_vector(8);
	state->counter3 = new_bit_vector(4);
	/* set them all counters to 0 */
	set_all_bits(state->counter1, FALSE);
	set_all_bits(state->counter2, FALSE);
	set_all_bits(state->counter3, FALSE);
	/* initialize the state machines */
	state->state_of_counter_1 = RESET;
	state->state_of_counter_2 = RESET;
	state->state_of_counter_3 = RESET;
}

/*---------------------------------------------------------------------------------------------
 * (function: free_glue_state)
 *-------------------------------------------------------------------------------------------*/
void free_glue_state(gl_device_state_t *state)
{
	free_bit_vector(state->counter1);
	free_bit_vector(state->counter2);
	free_bit_vector(state->counter3);
}

/*---------------------------------------------------------------------------------------------
 * (function: gl_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void gl_process_event_clump(LINKED_LIST *output_timeline, gl_device_state_t *state, event_t *clump_to_process)
{
	int i, j; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* now go through all the events in the current clump which represents a timestep */
	for (j = 0; j < clump_to_process->num_resources; j++)
	{	
		char *name = clump_to_process->resources[j]->name;
		int num_values = clump_to_process->resources[j]->num_values;
		char **value = clump_to_process->resources[j]->value;
		short to_free;

		/* need to get theresource name for dfault names */
		if (clump_to_process->resources[j]->name != NULL)
		{
			name = clump_to_process->resources[j]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[j]->default_name);
			to_free = TRUE;
		}

		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
				/* set them all counters to 0 */
				set_all_bits(state->counter1, FALSE);
				set_all_bits(state->counter2, FALSE);
				set_all_bits(state->counter3, FALSE);
				/* initialize the state machines */
				state->state_of_counter_1 = RESET;
				state->state_of_counter_2 = RESET;
				state->state_of_counter_3 = RESET;

				state->bus = 0;
				state->idle = 0;
			}
			else if (strcmp(value[0], "0") == 0)
			{
				/* initialize the state machines */
				state->state_of_counter_1 = STATE_IDLE;
				state->state_of_counter_2 = STATE_IDLE;
				state->state_of_counter_3 = STATE_IDLE;
			}
		}
		else if (strcmp(name, "idle") == 0)
		{
			if (value[0][0] == '1')
			{
				if (state->state_of_counter_1 == WAIT)
					state->state_of_counter_1 = STATE_IDLE;
				if (state->state_of_counter_2 == WAIT)
					state->state_of_counter_2 = STATE_IDLE;
				if (state->state_of_counter_3 == WAIT)
					state->state_of_counter_3 = STATE_IDLE;
			}
			else
			{
				if (state->state_of_counter_1 == STATE_IDLE)
					state->state_of_counter_1 = WAIT;
				if (state->state_of_counter_2 == STATE_IDLE)
					state->state_of_counter_2 = WAIT;
				if (state->state_of_counter_3 == STATE_IDLE)
					state->state_of_counter_3 = WAIT;
			}
		}
		else if (strcmp(name, "bus") == 0)
		{
			int bus_value = convert_bit_string_msb_to_int(&(value[0][0]), 8);

			if ((state->state_of_counter_1 == WAIT) || (state->state_of_counter_1 == PRO_FAST) ||(state->state_of_counter_1 == PRO_SLOW))
			{
				if (bus_value == 1)
				{
					/* state 1 = WAIT */
					state->state_of_counter_1 = WAIT;
				}
				else if (bus_value == 2)
				{
					/* state 1 = Pro Fast */
					state->state_of_counter_1 = PRO_FAST;
				}
				else if (bus_value == 3)
				{
					/* state 1 = Pro Slow */
					state->state_of_counter_1 = PRO_SLOW;
				}
			}
			if ((state->state_of_counter_2 == WAIT) || (state->state_of_counter_2 == PRO_FAST) ||(state->state_of_counter_2 == PRO_SLOW))
			{
				if (bus_value == 5)
				{
					/* state 2 = WAIT */
					state->state_of_counter_2 = WAIT;
				}
				else if (bus_value == 6)
				{
					/* state 2 = Pro Fast */
					state->state_of_counter_2 = PRO_FAST;
				}
				else if (bus_value == 7)
				{
					/* state 2 = Pro Slow */
					state->state_of_counter_2 = PRO_SLOW;
				}
			}
			if ((state->state_of_counter_3 == WAIT) || (state->state_of_counter_3 == PRO_FAST) ||(state->state_of_counter_3 == PRO_SLOW))
			{
				if (bus_value == 13)
				{
					/* state 3 = WAIT */
					state->state_of_counter_3 = WAIT;
				}
				else if (bus_value == 14)
				{
					/* state 3 = Pro Fast */
					state->state_of_counter_3 = PRO_FAST;
				}
				else if (bus_value == 15)
				{
					/* state 3 = Pro Slow */
					state->state_of_counter_3 = PRO_SLOW;
				}
			}
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	/* check if the state has changed.  If it has check if we need to output something */
	if (state->old_state_of_counter_1 != state->state_of_counter_1)
	{
		output_time_used = send_outputs_depending_on_state(output_time, output_timeline, output_time_used, state->state_of_counter_1, 1);
	}
	if (state->old_state_of_counter_2 != state->state_of_counter_2)
	{
		output_time_used = send_outputs_depending_on_state(output_time, output_timeline, output_time_used, state->state_of_counter_2, 2);
	}
	if (state->old_state_of_counter_3 != state->state_of_counter_3)
	{
		output_time_used = send_outputs_depending_on_state(output_time, output_timeline, output_time_used, state->state_of_counter_3, 3);
	}

	/* record the state for next pass */
	state->old_state_of_counter_1 = state->state_of_counter_1;
	state->old_state_of_counter_2 = state->state_of_counter_2;
	state->old_state_of_counter_3 = state->state_of_counter_3;

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: send_outputs_depending_on_state)
 *-------------------------------------------------------------------------------------------*/
short send_outputs_depending_on_state(mpz_t *output_time, LINKED_LIST *output_timeline, short output_time_used, short state, int counter)
{
	char *resource_name;
	char ** output_values;

	if (state == WAIT)
	{
		/* put the data on the output */
		/* create the resource name */
		resource_name = (char*)malloc(sizeof(char)*9);
		sprintf(resource_name, "counter%d", counter);
		/* create the character array */
		output_values = (char**)malloc(sizeof(char*)*1);
		output_values[0] = (char*)malloc(sizeof(char)*11);
		sprintf(output_values[0], "count stop"); 
		/* add this event to the output list since this results is an output event */
		add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
		
		return TRUE;
	}
	else if (state == PRO_FAST)
	{
		/* put the data on the output */
		/* create the resource name */
		resource_name = (char*)malloc(sizeof(char)*9);
		sprintf(resource_name, "counter%d", counter);
		/* create the character array */
		output_values = (char**)malloc(sizeof(char*)*1);
		output_values[0] = (char*)malloc(sizeof(char)*11);
		sprintf(output_values[0], "count fast"); 
		/* add this event to the output list since this results is an output event */
		add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
		
		return TRUE;
	}
	else if (state == PRO_SLOW)
	{
		/* put the data on the output */
		/* create the resource name */
		resource_name = (char*)malloc(sizeof(char)*9);
		sprintf(resource_name, "counter%d", counter);
		/* create the character array */
		output_values = (char**)malloc(sizeof(char*)*1);
		output_values[0] = (char*)malloc(sizeof(char)*11);
		sprintf(output_values[0], "count slow"); 
		/* add this event to the output list since this results is an output event */
		add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
		
		return TRUE;
	}

	return output_time_used;
}
