/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_lziv.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

lz_device_state_t lz_state;
lz_device_state_t output_state;
static decompress_first_pass;

void initialize_lziv_state(lz_device_state_t *state);
void free_lziv_state(lz_device_state_t *state);
void lz_process_event_clump(LINKED_LIST *output_timeline, lz_device_state_t *state, event_t *clump_to_process);
void initialize_the_code_book(lz_device_state_t *state);
short lziv_search_for_code_word(lz_device_state_t *state, short last_byte);

void lziv_decompress(lz_device_state_t *state, short force);

void debug_the_code_book(lz_device_state_t *state);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_lziv)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_lziv(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_lziv_state(&lz_state);
	initialize_lziv_state(&output_state);
	lz_state.workload_num = workload->workload_number;
	decompress_first_pass = TRUE;

	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		lz_process_event_clump(output_event_timeline, &lz_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_lziv_state(&lz_state);
	free_lziv_state(&output_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_lziv_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_lziv_state(lz_device_state_t *state)
{
	state->input_data = new_bit_vector(8);
	state->input_block_data = NULL;
	state->num_input_block_data = 0;

	state->stream = new_bit_vector(4096); // max codebook size for all 0 or all 1 entry
	state->bits_in_stream = 0;

	state->output_data = new_bit_vector(12);

	/* make the initial code book */
	initialize_the_code_book(state);

	state->is_input_block = -99;
}

/*---------------------------------------------------------------------------------------------
 * (function: free_lziv_state)
 *-------------------------------------------------------------------------------------------*/
void free_lziv_state(lz_device_state_t *state)
{
	int i;

	if (state->input_data != NULL)
		free_bit_vector(state->input_data);
	if (state->stream != NULL)
		free_bit_vector(state->stream);
	if (state->output_data != NULL)
		free_bit_vector(state->output_data);

	for (i = 0; i < state->size_code_book; i++)
	{
		free_bit_vector(state->code_book[i]);
	}
	free(state->code_book);
}

/*---------------------------------------------------------------------------------------------
 * (function: lz_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void lz_process_event_clump(LINKED_LIST *output_timeline, lz_device_state_t *state, event_t *clump_to_process)
{
	int i, index_of_resource, k; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;
	short is_output = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* now go through all the events in the current clump which represents a timestep */
	for (index_of_resource = 0; index_of_resource < clump_to_process->num_resources; index_of_resource++)
	{	
		char *name = clump_to_process->resources[index_of_resource]->name;
		int num_values = clump_to_process->resources[index_of_resource]->num_values;
		char **value = clump_to_process->resources[index_of_resource]->value;
		short to_free;

		/* need to get theresource name for dfault names */
		if (clump_to_process->resources[index_of_resource]->name != NULL)
		{
			name = clump_to_process->resources[index_of_resource]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[index_of_resource]->default_name);
			to_free = TRUE;
		}

	
		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
				state->is_input_block = -99;
			}
			else if (strcmp(value[0], "0") == 0)
			{
			}
		}
		else if (strcmp(name, "input_block") == 0)
		{
			/* IF - input data available then put it on the buffer */
			short next_short;

			assert(state->num_input_block_data == 0);

			state->top_input_block_data = 0;

			state->num_input_block_data = strlen(value[0])/2;
			state->input_block_data = (bit_vector **)malloc(sizeof(bit_vector*)*state->num_input_block_data);

			for (i = 0; i < state->num_input_block_data; i++)
			{
				state->input_block_data[i] = new_bit_vector(8);
				next_short = convert_hex_string_to_short(&(value[0][i*2]), 2);
				set_bits_with_short_at(state->input_block_data[i], 0, next_short);
			}

			printf("\nlziv reverse %d:\n", lz_state.workload_num); 

			assert(state->is_input_block == -99);
			state->is_input_block = TRUE;
		}
		else if (strcmp(name, "input_data_valid") == 0)
		{
			/* If - the input data is valid and we are ready top process.  This is a simplified view. */
			char **output_values;

			assert(state->is_input_block == TRUE || state->is_input_block == FALSE);
			
			if (state->is_input_block == TRUE)
			{
				/* process next piece of block */
				copy_n_bits_at_x_to_y(8, state->input_block_data[state->top_input_block_data], 0, state->input_data, 0);
				state->top_input_block_data ++;
			}

			is_output = lziv_search_for_code_word(state, FALSE);
	
			if (is_output == TRUE)
			{
				/* put the data on the output */
		
				/* create the resource name */
				resource_name = (char*)malloc(sizeof(char)*10);
				sprintf(resource_name, "output");
		
				/* create the character array */
				output_values = (char**)malloc(sizeof(char*)*1);
				output_values[0] = convert_bitvector_to_hex_string(state->output_data);
		
				/* add this event to the output list since this results is an output event */
				add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
				output_time_used = TRUE;

				/* for each ouput chracter do a lziv decode sequence to reconstruct the stream to verify working correctly */
				lziv_decompress(state, FALSE);
			}
		}
		else if (strcmp(name, "input_termination") == 0)
		{
			/* If - the input data is valid and we are ready top process.  This is a simplified view. */
			char **output_values;

			assert(state->is_input_block == TRUE || state->is_input_block == FALSE);

			/* reset the block data */
			for (i = 0; i < state->num_input_block_data; i++)
			{
				free_bit_vector(state->input_block_data[i]);
			}
			free(state->input_block_data);
			state->input_block_data = NULL;
			state->num_input_block_data = 0;

			state->is_input_block = -99;

			if (state->bits_in_stream > 8)
			{
				is_output = lziv_search_for_code_word(state, TRUE);
			}
	
			if (is_output == TRUE)
			{
				/* put the data on the output */
		
				/* create the resource name */
				resource_name = (char*)malloc(sizeof(char)*10);
				sprintf(resource_name, "output");
		
				/* create the character array */
				output_values = (char**)malloc(sizeof(char*)*1);
				output_values[0] = convert_bitvector_to_hex_string(state->output_data);
		
				/* add this event to the output list since this results is an output event */
				add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 
			}
			lziv_decompress(state, TRUE);

			output_time_used = TRUE;

			/* now generate the end signal */
			/* create the resource name */
			resource_name = (char*)malloc(sizeof(char)*10);
			sprintf(resource_name, "output");
	
			/* create the character array */
			output_values = (char**)malloc(sizeof(char*)*1);
			output_values[0] = (char*)malloc(sizeof(char)*13);
			sprintf(output_values[0], "111111111111");
	
			/* add this event to the output list since this results is an output event */
			add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 

			/* now generate the tail end bits from the bit stream in the form of first 4 bits indicat num tail bits
			 * and the remaining bits are the value */ 
			/* create the resource name */
			resource_name = (char*)malloc(sizeof(char)*10);
			sprintf(resource_name, "output");
	
			/* create the character array */
			output_values = (char**)malloc(sizeof(char*)*1);
			output_values[0] = (char*)malloc(sizeof(char)*13);
			/* PAJ */
			sprintf(output_values[0], "%b10101010", state->bits_in_stream);
	
			/* add this event to the output list since this results is an output event */
			add_event_to_timeline(output_timeline, output_time, resource_name, output_values, 1, OUTPUT_EVENT); 

			printf("\n\n");

			/* reset for the next block */
			free_lziv_state(&lz_state);
			free_lziv_state(&output_state);
			initialize_lziv_state(&lz_state);
			initialize_lziv_state(&output_state);
			decompress_first_pass = TRUE;
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: lziv_search_for_code_word)
 * Compression function
 *-------------------------------------------------------------------------------------------*/
short lziv_search_for_code_word(lz_device_state_t *state, short last_byte)
{
	int i, size_bit_vector;

	bit_vector *temp_bit_stream = new_bit_vector(4104); // = 4096 + 8 initial bits
	int size_of_temp_bit_stream = 0;
	int current_code = -1;
	int current_code_size = 0;
	short found_max_sized_code = FALSE;
	short found_output = FALSE;
	short found_match_this_pass;

	/* copy in the input data to the stream */	
	copy_n_bits_at_x_to_y(8, state->input_data, 0, state->stream, state->bits_in_stream);
	state->bits_in_stream += 8;

	/* start with 8 bits */	
	copy_n_bits_at_x_to_y(8, state->stream, 0, temp_bit_stream, 0);
	size_of_temp_bit_stream = 8;

	for (size_bit_vector = 8; size_bit_vector < state->bits_in_stream; size_bit_vector++)
	{
		found_match_this_pass = FALSE;
		/* search for this code in the codebook */
		for (i = 0; i < state->size_code_book; i++)
		{
			if ((size_of_temp_bit_stream == state->code_book[i]->num_bits) &&
				(bits_are_equal_n_bits (state->code_book[i], temp_bit_stream, size_of_temp_bit_stream) == TRUE))
			{
				/* IF - there is a code for this entry record and try a larger bitstream */
				current_code = i;
				current_code_size = size_of_temp_bit_stream;
				found_match_this_pass = TRUE;
				break;
			}
		}

		if (found_match_this_pass != TRUE)
		{
			/* We've found the biggest match */
			found_max_sized_code = TRUE;
			break;
		}

		if (size_bit_vector != state->bits_in_stream - 1)
		{
			/* make the stream to search for in the codebook 1 bit longer */
			set_bit(temp_bit_stream, size_bit_vector, get_bit(state->stream, size_bit_vector));
			size_of_temp_bit_stream ++;
		}
	}

	if (((state->size_code_book == CODE_BOOK_SIZE) && (found_max_sized_code == TRUE)) || (last_byte == TRUE))
	{
		/* IF - the codebook is full then we just simply return the biggest code found, adjust the stream accordingly */
		/* IF - this is the last byte in the stream, then we send the last coded word, put a termination sequence, and then
		 * a few bits for the spillover */

		/* set the output vector */
		set_all_bits(state->output_data, FALSE);
		set_n_bits_with_int(state->output_data, current_code, 12);

		/* adjust the stream */
		copy_n_bits_at_x_to_y(state->bits_in_stream, state->stream, 0, temp_bit_stream, 0);
		copy_n_bits_at_x_to_y(state->bits_in_stream - current_code_size, temp_bit_stream, current_code_size, state->stream, 0);
		state->bits_in_stream = state->bits_in_stream - current_code_size;

		found_output = TRUE;
	}
	else if (found_max_sized_code == TRUE)
	{
		/* ELSE - if the code book isn't full, and we found a max sized code, then add a new code, output the current symbol,
		 * and update the stream accordingly */

		/* add to code book */
		state->code_book[state->size_code_book] = new_bit_vector(current_code_size+1);
		copy_n_bits_at_x_to_y(current_code_size+1, state->stream, 0, state->code_book[state->size_code_book], 0);
		state->size_code_book++;

		/* set the output vector */
		set_all_bits(state->output_data, FALSE);
		set_n_bits_with_int(state->output_data, current_code, 12);

		/* adjust the stream */
		copy_n_bits_at_x_to_y(state->bits_in_stream, state->stream, 0, temp_bit_stream, 0);
		copy_n_bits_at_x_to_y(state->bits_in_stream - current_code_size, temp_bit_stream, current_code_size, state->stream, 0);
		state->bits_in_stream = state->bits_in_stream - current_code_size;

		found_output = TRUE;
	}

	free_bit_vector(temp_bit_stream);

	return found_output;
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_the_code_book)
 *-------------------------------------------------------------------------------------------*/
void initialize_the_code_book(lz_device_state_t *state)
{
	int i;
	int index = 0;
	bit_vector *temp_vector = new_bit_vector(INITIAL_CODE_BOOK);

	state->code_book = (bit_vector**)malloc(sizeof(bit_vector*)*CODE_BOOK_SIZE);
	state->size_code_book = 0;

	int number_of_symbols = my_power(2, INITIAL_CODE_BOOK);

	for (i = 0; i < number_of_symbols; i++) 
	{
		state->code_book[index] = new_bit_vector(INITIAL_CODE_BOOK);
		/* Note that the set bits with short puts it in order 3 = LSIndex[0] -> 00000011 <- MSIndex[7]
		 * this is why we do the Reverse_bits */
		set_bits_with_short_at(state->code_book[index], 0, i);
		reverse_bits(state->code_book[index]);

		index ++;
		state->size_code_book ++;
	}

	free_bit_vector(temp_vector);
}

/*---------------------------------------------------------------------------------------------
 * (function: lziv_decompress)
 *-------------------------------------------------------------------------------------------*/
void lziv_decompress(lz_device_state_t *state, short force)
{
	static bit_vector *old_vector = NULL;
	static bit_vector *new_vector = NULL;
	static bit_vector *output_stream = NULL;
	static bit_vector *C_val = NULL;
	static int current_spot_in_output_stream;
	int index_code_table_old;
	int index_code_table_new;
	bit_vector *S;
	bit_vector *temp_bit_stream = new_bit_vector(4104); // = 4096 + 8 initial bits

	if (decompress_first_pass == TRUE)
	{
		/* 2    OLD = first input code
		 * 3    output translation of OLD
		 */
		if (old_vector == NULL)
		{
			/* don't reinitialize */
			old_vector = new_bit_vector(12);
			new_vector = new_bit_vector(12);
			output_stream = new_bit_vector(4200);
			C_val = new_bit_vector(1);
		}
		current_spot_in_output_stream = 0;

		copy_n_bits_at_x_to_y(12, state->output_data, 0, old_vector, 0);
//		reverse_bits(old_vector);
		index_code_table_old = get_int_from_bit_vector(old_vector);
		copy_n_bits_at_x_to_y(output_state.code_book[index_code_table_old]->num_bits, output_state.code_book[index_code_table_old], 0, output_stream, current_spot_in_output_stream);
		current_spot_in_output_stream += output_state.code_book[index_code_table_old]->num_bits;
		
		decompress_first_pass = FALSE;
	}
	else
	{
		/* 4	WHILE not end of input stream
		 * 5   		NEW = next input code
		 * 6   		IF NEW is not in the string table
		 * 7           		S = translation of OLD
		 * 8			S = S + C_val
		 * 9		ELSE
		 * 10			S = translation of NEW
		 * 11		output S
		 * 12		C_val = first character of S
		 * 13		OLD + C_val to the string table
		 * 14 		OLD = NEW
		 * 15	END WHILE
		 * */
		copy_n_bits_at_x_to_y(12, state->output_data, 0, new_vector, 0);
		index_code_table_new = get_int_from_bit_vector(new_vector);
		index_code_table_old = get_int_from_bit_vector(old_vector);

		if (index_code_table_new >= output_state.size_code_book)
		{
			S = new_bit_vector(output_state.code_book[index_code_table_old]->num_bits+1);
			copy_n_bits_at_x_to_y(output_state.code_book[index_code_table_old]->num_bits, output_state.code_book[index_code_table_old], 0, S, 0);
			/* 8 */
			set_bit(S, output_state.code_book[index_code_table_old]->num_bits, get_bit(C_val, 0));
		}
		else
		{
			S = new_bit_vector(output_state.code_book[index_code_table_new]->num_bits);
			copy_n_bits_at_x_to_y(output_state.code_book[index_code_table_new]->num_bits, output_state.code_book[index_code_table_new], 0, S, 0);
		}
		
		/* 11 */
		copy_n_bits_at_x_to_y(S->num_bits, S, 0, output_stream, current_spot_in_output_stream);
		current_spot_in_output_stream += S->num_bits;

		/* 12 */
		set_bit(C_val, 0, get_bit(S, 0));

		/* 13 */
		if (output_state.size_code_book < CODE_BOOK_SIZE)
		{
			output_state.code_book[output_state.size_code_book] = new_bit_vector(output_state.code_book[index_code_table_old]->num_bits+1);
			copy_n_bits_at_x_to_y(output_state.code_book[index_code_table_old]->num_bits, output_state.code_book[index_code_table_old], 0, output_state.code_book[output_state.size_code_book], 0);
			set_bit(output_state.code_book[output_state.size_code_book], output_state.code_book[index_code_table_old]->num_bits, get_bit(C_val, 0));
			output_state.size_code_book++;
		}

		/* 14 */
		copy_n_bits_at_x_to_y(12, new_vector, 0, old_vector, 0);

		free_bit_vector(S);
	}

	/* now look at the output stream and decide if we need to */
	while (current_spot_in_output_stream >= 8) 
	{
		short value = get_short_from_bits_at(output_stream, 0, 8);

		// need an extra case since for every two hex characters a leading 0 won't be printed
		if(value < 16)
			printf("0%x", get_short_from_bits_at(output_stream, 0, 8));
		else
			printf("%x", get_short_from_bits_at(output_stream, 0, 8));

		fflush(stdout);
		/* adjust the stream */
		copy_n_bits_at_x_to_y(current_spot_in_output_stream, output_stream, 0, temp_bit_stream, 0);
		copy_n_bits_at_x_to_y(current_spot_in_output_stream - 8, temp_bit_stream, 8, output_stream, 0);
		current_spot_in_output_stream = current_spot_in_output_stream - 8;
	}

	free_bit_vector(temp_bit_stream);
}

/*---------------------------------------------------------------------------------------------
 * (function: debug_the_code_book)
 *-------------------------------------------------------------------------------------------*/
void debug_the_code_book(lz_device_state_t *state)
{
	int i;

	for (i = 0; i < state->size_code_book; i++) 
	{
		printf("index: %d code->", i);
		fprint_bits_out (stdout, state->code_book[i], fprint_bit);
		printf("\n");
		fflush(stdout);
	}
}

