/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "fm_portkey.h"
#include "fm_simulation_utils.h"
#include "workload_utils.h"

pe_device_state_t keypad_state;

void initialize_keypad_state(pe_device_state_t *state);
void free_keypad_state(pe_device_state_t *state);

void process_event_clump(LINKED_LIST *output_timeline, pe_device_state_t *state, event_t *clump_to_process);
short check_for_portkey_output_updates(pe_device_state_t *state, mpz_t *output_time, LINKED_LIST *output_timeline, short current_value_of_output_time_used);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_keypad)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_keypad(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events->num_events;
	event_t *next_clump;
	workload_t output;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_keypad_state(&keypad_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		process_event_clump(output_event_timeline, &keypad_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);
	}

	/* now that all the events are processed, create the output file */
	generate_an_output_workload_file(output_file, &output, workload->design_name, workload->workload_number);

	/* cleanup */
	free_keypad_state(&keypad_state);
	free_workload(&output);
	free(output.design_name);
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_keypad_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_keypad_state(pe_device_state_t *state)
{
	/* initialize all the state registers */
	state->control_register = new_bit_vector(16);
	state->direction_register = new_bit_vector(16);
	state->interrupt_register = new_bit_vector(16);
	state->value_register = new_bit_vector(16);

	/* create the keypad fifo */
	state->fifo_index = 0;
	state->fifo_size = 0;
	state->fifo_list = (char **)malloc(sizeof(char*)*FIFO_MAX_SIZE);
}

/*---------------------------------------------------------------------------------------------
 * (function: free_keypad_state)
 *-------------------------------------------------------------------------------------------*/
void free_keypad_state(pe_device_state_t *state)
{
	/* initialize all the state registers */
	free_bit_vector(state->control_register);
	free_bit_vector(state->direction_register);
	free_bit_vector(state->interrupt_register);
	free_bit_vector(state->value_register);
}

/*---------------------------------------------------------------------------------------------
 * (function: process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void process_event_clump(LINKED_LIST *output_timeline, pe_device_state_t *state, event_t *clump_to_process)
{
	int i, j; 
	char *resource_name;
	int presses_detected = 0;
	short *column_presses = NULL;
	short *row_presses = NULL;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;
	short interrupt1 = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add(*output_time, *output_time, *environment.minimum_operating_speed); // add fast clock speed

	/* setup the data structure to detect key presses */
	column_presses = (short *)malloc(sizeof(int)*8);
	row_presses = (short *)malloc(sizeof(int)*8);
	for (i = 0; i < 8; i++)
	{
		column_presses[i] = FALSE;
		row_presses[i] = FALSE;
	}

	/* now go through all the events in the current clump which represents a timestep */
	for (j = 0; j < clump_to_process->num_resources; j++)
	{	
		char *name = clump_to_process->resources[j]->name;
		int num_values = clump_to_process->resources[j]->num_values;
		char **value = clump_to_process->resources[j]->value;
		short to_free;

		/* need to get theresource name for dfault names */
		if (clump_to_process->resources[j]->name != NULL)
		{
			name = clump_to_process->resources[j]->name;
			to_free = FALSE;
		}
		else
		{
			name = get_default_resource_name(clump_to_process->resources[j]->default_name);
			to_free = TRUE;
		}

	
		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
				/* assumption here is the 1 is a reset */
				set_all_bits(state->control_register, FALSE);
				set_all_bits(state->direction_register, FALSE);
				set_all_bits(state->interrupt_register, FALSE);
				set_all_bits(state->value_register, FALSE);
			}
			else if (strcmp(value[0], "0") == 0)
			{
				output_time_used = check_for_portkey_output_updates(state, output_time, output_timeline, output_time_used);
			}
		}
		else if (strcmp(name, "gpio_pin") == 0)
		{
			/* IF - a reading GPIO_PIN sees a value should check if change */
			short pin_number = (short)atoi(value[0]);
			short bit_value = (short)atoi(value[1]);
			short bit_boolean = bit_value ? TRUE : FALSE;
			digital_bit mask = get_bit (state->direction_register, pin_number);
			digital_bit interrupt = get_bit (state->interrupt_register, pin_number);
			digital_bit current_value = get_bit (state->value_register, pin_number);
			
			if ((current_value != bit_boolean) && (mask == FALSE))
			{
				/* IF this is a change of value then write it */
				set_bit(state->value_register, pin_number, bit_boolean);

				if (interrupt == TRUE && interrupt1 == FALSE)
				{
					/* IF we are supposed to send interrupts on this bit */
					resource_name = (char*)malloc(sizeof(char)*15);
					sprintf(resource_name, "interrupt1");

					interrupt1 = TRUE;

					/* adds keypresses 1us per interrupt */
					output_time_used = TRUE;
					/* add this event to the output list since this results in an output event */
					add_event_to_timeline(output_timeline, output_time, resource_name, NULL, 0, OUTPUT_EVENT); 
				}
			}
		}
		else if (strcmp(name, "keypad") == 0)
		{
			/* IF - a keypad press happens on this cycle */

			/* add the row and column press to the keypress queu */
			/* these will be processed at the end of all events at this time */
			column_presses[atoi(value[0])] = TRUE;
			row_presses[atoi(value[1])] = TRUE;
		}
		else if (strcmp(name, "spi") == 0)
		{
			int address = convert_bit_string_lsb_to_int(&(value[0][1]), 7);
			unsigned short first_read_bit;
			unsigned short second_read_bit;

			if (value[0][0] == '1')
			{
				/* If - this is an SPI write */
				unsigned short first_write_bit = (unsigned short)convert_bit_string_lsb_to_int(&(value[0][8]), 8);
				unsigned short second_write_bit =(unsigned short)convert_bit_string_lsb_to_int(&(value[0][16]), 8);
				
				if (address == 0)
				{
					/* write to control register */
					set_bits_with_short_at(state->control_register, 0, first_write_bit);
					set_bits_with_short_at(state->control_register, 8, second_write_bit);
				}
				else if (address == 1)
				{
					/* write to Interrupt register */
					set_bits_with_short_at(state->interrupt_register, 0, first_write_bit);
					set_bits_with_short_at(state->interrupt_register, 8, second_write_bit);
				}
				else if (address == 2)
				{
					/* write to Direction register */
					set_bits_with_short_at(state->direction_register, 0, first_write_bit);
					set_bits_with_short_at(state->direction_register, 8, second_write_bit);
				}
				else if (address == 3)
				{
					/* need to mask here with direction register */
					first_read_bit = get_short_from_bits_at(state->direction_register, 0, 8);
					second_read_bit = get_short_from_bits_at(state->direction_register, 8, 8);

					/* masking */
					set_bits_with_short_with_mask_at(state->value_register, 0, first_write_bit, first_read_bit);
					set_bits_with_short_with_mask_at(state->value_register, 8, second_write_bit, second_read_bit);

					/* output changes in the value register regardless of if there's a change or not */
					output_time_used = check_for_portkey_output_updates(state, output_time, output_timeline, output_time_used);
				}
			}
			else
			{
				/* Else - this is an SPI read */
				int string_value;
				char **spi_out;
				spi_out = (char**)malloc(sizeof(char*)*1);

				if (address == 0)
				{
					/* read to control register */
					first_read_bit = get_short_from_bits_at(state->control_register, 0, 8);
					second_read_bit = get_short_from_bits_at(state->control_register, 8, 8);
					/* only do for addresses 0 through 3 */
					string_value = first_read_bit + (second_read_bit << 8);
					spi_out[0] = convert_int_to_bit_string(string_value, 16);
				}
				else if (address == 1)
				{
					/* read to Interrupt register */
					first_read_bit = get_short_from_bits_at(state->interrupt_register, 0, 8);
					second_read_bit = get_short_from_bits_at(state->interrupt_register, 8, 8);
					/* only do for addresses 0 through 3 */
					string_value = first_read_bit + (second_read_bit << 8);
					spi_out[0] = convert_int_to_bit_string(string_value, 16);
				}
				else if (address == 2)
				{
					/* read to Direction register */
					first_read_bit = get_short_from_bits_at(state->direction_register, 0, 8);
					second_read_bit = get_short_from_bits_at(state->direction_register, 8, 8);
					/* only do for addresses 0 through 3 */
					string_value = first_read_bit + (second_read_bit << 8);
					spi_out[0] = convert_int_to_bit_string(string_value, 16);
				}
				else if (address == 3)
				{
					/* read to Value register */
					first_read_bit = get_short_from_bits_at(state->value_register, 0, 8);
					second_read_bit = get_short_from_bits_at(state->value_register, 8, 8);
					/* only do for addresses 0 through 3 */
					string_value = first_read_bit + (second_read_bit << 8);
					spi_out[0] = convert_int_to_bit_string(string_value, 16);
				}
				else if (address == 4)
				{
					/* This is the fifo of keypresses */
					if (state->fifo_size > 0)
					{
						/* IF - there are values to return */
						spi_out[0] = state->fifo_list[state->fifo_index];

						/* change the index (needs to wrap) and the size of the fifo since the top is read */
						state->fifo_index = state->fifo_index == FIFO_MAX_SIZE-1 ? 0 : state->fifo_index + 1;
						state->fifo_size --;
					}
				}

				/* create the resource name */
				resource_name = (char*)malloc(sizeof(char)*10);
				sprintf(resource_name, "spi_SO");

				/* add this event to the output list since this results is an output event */
				add_event_to_timeline(output_timeline, output_time, resource_name, spi_out, 1, OUTPUT_EVENT); 
				output_time_used = TRUE;

				/* since add_event_to_timeline makes own internal array can free spi_out */
				free(spi_out);
			}
		}

		if (to_free == TRUE)
		{
			free(name);
		}
	}

	/* cleanup the clump */
	free_event_clump(clump_to_process);

	/* process all the keypresses to add to the fifo */
	/* assumption is that the columns send high outputs on the poll, going from lsb to msb */
	/* simulate ghosting by looking at each on column and for each row element on at that time we get a keypress */
	for (i = 0; i < 8; i++)
	{
		if (column_presses[i] == TRUE)
		{
			for (j = 0; j < 8; j++)
			{
				if (row_presses[j] == TRUE)
				{
					/* generate a fifo entry */
					int value = i + (j << 8);
					char *bit_string_of_press = convert_int_to_bit_string(value, 16);

					/* store in fifo list */
					if (state->fifo_size < FIFO_MAX_SIZE)		
					{
						/* IF - there is room in the list...otherwise it is lost */
						int add_index;
						mpz_t *time_to_add;

						/* update the fifo access */
						state->fifo_size++;
						add_index = (state->fifo_index+state->fifo_size > FIFO_MAX_SIZE) ? (state->fifo_size-1) - (FIFO_MAX_SIZE - state->fifo_index) : state->fifo_index+state->fifo_size-1; 
						state->fifo_list[add_index] = bit_string_of_press;

						/* generate an interrupt */
						presses_detected++; // each interrupt comes after some time

						/* create the resource name */
						resource_name = (char*)malloc(sizeof(char)*15);
						sprintf(resource_name, "interrupt0");

						time_to_add = (mpz_t*)malloc(sizeof(mpz_t));

						mpz_init(*time_to_add);
						/* adds keypresses 1us per interrupt ... maybe the file should define this? */
						mpz_mul_ui(*time_to_add, *environment.minimum_operating_speed, (unsigned int)(10*presses_detected));
						mpz_add(*time_to_add, *output_time, *time_to_add);
						output_time_used = TRUE;
						/* add this event to the output list since this results in an output event */
						add_event_to_timeline(output_timeline, time_to_add, resource_name, NULL, 0, OUTPUT_EVENT); 
					}
				}
			}
		}
	}

	free(column_presses);
	free(row_presses);

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: check_for_portkey_output_updates)
 * When the value register is written to we assume that the outputs need to all be generated,
 * so we go through the list and update these outputs
 *-------------------------------------------------------------------------------------------*/
short check_for_portkey_output_updates(pe_device_state_t *state, mpz_t *output_time, LINKED_LIST *output_timeline, short current_value_of_output_time_used)
{
	int i;
	digital_bit direction_mask;
	digital_bit current_value;
	char **char_values;
	char *resource_name;
	short output_time_used = current_value_of_output_time_used;

	/* ELSE - any output GPIOs need to send out there value */
	for (i = 0; i < 16; i++)
	{
		direction_mask = get_bit (state->direction_register, i);
		current_value = get_bit (state->value_register, i);

		if (direction_mask == TRUE)
		{
			/* create the resource name */
			resource_name = (char*)malloc(sizeof(char)*10);
			sprintf(resource_name, "gpio_pin");

			char_values = (char **)malloc(sizeof(char*)*2);
			/* create the value strings */
			/* the pin number */
			char_values[0] = (char *)malloc(sizeof(char)*3);
			sprintf(char_values[0], "%d", i);
			/* create the value of the output */
			char_values[1] = (char *)malloc(sizeof(char)*2);
			if (current_value == TRUE)
			{
				sprintf(char_values[1], "1", i);
			}
			else
			{
				sprintf(char_values[1], "0", i);
			}

			/* add this event to the output list since this results is an output event */
			add_event_to_timeline(output_timeline, output_time, resource_name, char_values, 2, OUTPUT_EVENT); 
			output_time_used = TRUE;

			free(char_values);
		}
	}

	return output_time_used;
}
