/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "types.h"
#include "globals.h"
#include "util.h"
#include "gmp.h"

void read_top(xmlNode * a_node, workload_t *w);
void read_events(xmlNode * a_node, workload_t *w);
void read_conditions(xmlNode * a_node, workload_t *w);
void read_next_condition(xmlNode * a_node, workload_t *w);
void read_next_event(xmlNode * a_node, event_t **list_of_events, int event_index);
void read_next_resource(xmlNode * a_node, resource_t *resource_list);
void read_next_assertion(xmlNode * a_node, assert_t *assert_element);

/*---------------------------------------------------------------------------------------------
 * (function: read_workload_input)
 * This reads in pre created workload XML files.  Wave generation does this to create a file,
 * so this is only used in rare cases (and not tested that much I might add)
 *-------------------------------------------------------------------------------------------*/
workload_t *read_workload_input(char *file_name)
{
	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;
	workload_t *return_workload = (workload_t*)malloc(sizeof(workload_t));

	/* this initialize the library and check potential ABI mismatches between the version it was compiled for and the actual shared library used. */
	LIBXML_TEST_VERSION

	/*parse the file and get the DOM */
	doc = xmlParseFile(file_name);

	if (doc == NULL) 
	{
		printf("error: could not parse file LIB file\n");
	}

	/*Get the root element node */
	root_element = xmlDocGetRootElement(doc);

	/* parse in the XML to data structs */
	read_top(root_element, return_workload);

	/*free the document */
	xmlFreeDoc(doc);

	/* Free the global variables that may have been allocated by the parser.  */
	xmlCleanupParser();

	return return_workload;
}


/*---------------------------------------------------------------------------------------------
 * (function: read_top)
 *-------------------------------------------------------------------------------------------*/
void read_top(xmlNode * a_node, workload_t *w)
{
	int i, j;
	xmlNode *top_node = NULL;
	xmlNode *cur_node = NULL;
	xmlChar *number_tfe_string = NULL;

	for (top_node = a_node; top_node; top_node = top_node->next) 
	{
		if (strcmp(top_node->name, "workload") == 0)
		{
			/* IF - Make sure this is the top node */

			for (cur_node = top_node->children; cur_node; cur_node = cur_node->next) 
			{
				if (cur_node->type == XML_ELEMENT_NODE) 
				{
					/* IF - these top level nodes should be lements */
					if (strcmp(cur_node->name, "conditions") == 0)
					{
						read_conditions(cur_node->children, w);
					}
					else if (strcmp(cur_node->name, "events") == 0)
					{
						read_events(cur_node->children, w);
					}
					else if (strcmp(cur_node->name, "design_name") == 0)
					{
						w->design_name = get_text(cur_node->children);	
					}
					else if (strcmp(cur_node->name, "workload_number") == 0)
					{
						char *temp_string = get_text(cur_node->children);
						w->workload_number = atoi(temp_string);
						free(temp_string);
					}
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_conditions)
 *-------------------------------------------------------------------------------------------*/
void read_conditions(xmlNode * a_node, workload_t *w)
{
	xmlNode *cur_node = NULL;

	/* initialize the number of pairs */
	w->conditions = (condition_t*)malloc(sizeof(condition_t));
	w->conditions->num_if_resource_pairs = 0;
	w->conditions->ifs = NULL;
	w->conditions->then_resources = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "condition") == 0)
		{
			/* read in the resource value */
			read_next_condition(cur_node->children, w);
		}
	}
}
/*---------------------------------------------------------------------------------------------
 * (function: read_conditions)
 *-------------------------------------------------------------------------------------------*/
void read_next_condition(xmlNode * a_node, workload_t *w)
{
	xmlNode *cur_node = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "if") == 0)
		{
			xmlNode *if_node = NULL;

			/* increment another resource to the list */
			w->conditions->num_if_resource_pairs++;

			/* allocate another spot on the ifs of resources */
			w->conditions->ifs = (resource_t **)realloc(w->conditions->ifs, sizeof(resource_t*)*w->conditions->num_if_resource_pairs);

			for (if_node = cur_node->children; if_node; if_node = if_node->next) 
			{
				if (strcmp(if_node->name, "resource") == 0)
				{
					w->conditions->ifs[w->conditions->num_if_resource_pairs-1] = (resource_t*)malloc(sizeof(resource_t));
					/* read in the resource value */
					read_next_resource(if_node->children, w->conditions->ifs[w->conditions->num_if_resource_pairs-1]);
				}
			}
		}
		else if (strcmp(cur_node->name, "resource") == 0)
		{
			/* allocate another spot on the ifs of resources */
			w->conditions->then_resources = (resource_t **)realloc(w->conditions->then_resources, sizeof(resource_t*)*w->conditions->num_if_resource_pairs);

			w->conditions->then_resources[w->conditions->num_if_resource_pairs-1] = (resource_t*)malloc(sizeof(resource_t));
			/* read in the resource value */
			read_next_resource(cur_node->children, w->conditions->then_resources[w->conditions->num_if_resource_pairs-1]);
		}
		else if (strcmp(cur_node->name, "time") == 0)
		{
			char *temp_string = get_text(cur_node->children);

			/* allocate another spot on the ifs of resources */
			w->conditions->then_times = (mpz_t **)realloc(w->conditions->then_times, sizeof(mpz_t*)*w->conditions->num_if_resource_pairs);

			w->conditions->then_times[w->conditions->num_if_resource_pairs-1] = (mpz_t *)malloc(sizeof(mpz_t));
			/* read in the resource value */
			w->conditions->then_times[w->conditions->num_if_resource_pairs-1] = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_events)
 *-------------------------------------------------------------------------------------------*/
void read_events(xmlNode * a_node, workload_t *w)
{
	xmlNode *cur_node = NULL;

	/* initialize the number of events */
	w->events = (events_t*)malloc(sizeof(events_t));
	w->events->num_events = 0;
	w->events->events = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "event") == 0)
		{
			/* increment another event to the list */
			w->events->num_events++;

			/* allocate another spot on the ifs of resources */
			w->events->events = (event_t **)realloc(w->events->events, sizeof(event_t*)*w->events->num_events);

			/* read in the resource value */
			read_next_event(cur_node->children, w->events->events, w->events->num_events-1);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_next_event)
 *-------------------------------------------------------------------------------------------*/
void read_next_event(xmlNode * a_node, event_t **list_of_events, int event_index)
{
	xmlNode *cur_node = NULL;
	resource_t *resource_to_add;

	/* allocate the resource memory */
	list_of_events[event_index] = (event_t*)malloc(sizeof(event_t));
	list_of_events[event_index]->num_resources = 0;
	list_of_events[event_index]->resources = NULL;
	list_of_events[event_index]->type = INPUT_EVENT;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			/* convert all times into nanosecond */
			list_of_events[event_index]->time = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "resource") == 0)
		{
			/* increment another resource to the list */
			list_of_events[event_index]->num_resources ++;

			/* allocate another spot on the ifs of resources */
			list_of_events[event_index]->resources = (resource_t **)realloc(list_of_events[event_index]->resources, sizeof(resource_t*)*list_of_events[event_index]->num_resources);

			list_of_events[event_index]->resources[list_of_events[event_index]->num_resources-1] = (resource_t*)malloc(sizeof(resource_t));
			/* read in the resource value */
			read_next_resource(cur_node->children, list_of_events[event_index]->resources[list_of_events[event_index]->num_resources-1]);
		}
		else if (strcmp(cur_node->name, "assert") == 0)
		{
			/* increment another resource to the list */
			list_of_events[event_index]->num_assertions ++;

			/* allocate another spot on the ifs of resources */
			list_of_events[event_index]->assertions = (assert_t **)realloc(list_of_events[event_index]->assertions, sizeof(assert_t*)*list_of_events[event_index]->num_assertions);

			list_of_events[event_index]->assertions[list_of_events[event_index]->num_assertions-1] = (assert_t*)malloc(sizeof(assert_t));
			/* read in the resource value */
			read_next_assertion(cur_node->children, list_of_events[event_index]->assertions[list_of_events[event_index]->num_assertions-1]);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_next_assertion)
 *-------------------------------------------------------------------------------------------*/
void read_next_assertion(xmlNode * a_node, assert_t *assert_element)
{
	xmlNode *cur_node = NULL;

	/* allocate the assert_element memory */
	assert_element->num_resources = 0;
	assert_element->resources = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			assert_element->time = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		if (strcmp(cur_node->name, "resource") == 0)
		{
			/* increment another resource to the list */
			assert_element->num_resources ++;

			/* allocate another spot on the ifs of resources */
			assert_element->resources = (resource_t **)realloc(assert_element->resources, sizeof(resource_t*)*assert_element->num_resources);

			assert_element->resources[assert_element->num_resources-1] = (resource_t*)malloc(sizeof(resource_t));
			/* read in the resource value */
			read_next_resource(cur_node->children, assert_element->resources[assert_element->num_resources-1]);
		}
	}
}
/*---------------------------------------------------------------------------------------------
 * (function: read_next_resource)
 *-------------------------------------------------------------------------------------------*/
void read_next_resource(xmlNode * a_node, resource_t *resource)
{
	xmlNode *cur_node = NULL;

	/* allocate the resource memory */
	resource->default_name = -1;
	resource->num_values = 0;
	resource->value = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "name") == 0)
		{
			resource->name = get_text(cur_node->children);	
		}
		if (strcmp(cur_node->name, "value0") == 0)
		{
			resource->value = (char**)realloc(resource->value, sizeof (char*)*1);
			resource->num_values = 1;
			resource->value[0] = get_text(cur_node->children);	
		}
		if (strcmp(cur_node->name, "value1") == 0)
		{
			resource->value = (char**)realloc(resource->value, sizeof (char*)*2);
			resource->num_values = 2;
			resource->value[1] = get_text(cur_node->children);	
		}
		if (strcmp(cur_node->name, "value2") == 0)
		{
			resource->value = (char**)realloc(resource->value, sizeof (char*)*3);
			resource->num_values = 3;
			resource->value[2] = get_text(cur_node->children);	
		}
		if (strcmp(cur_node->name, "value3") == 0)
		{
			resource->value = (char**)realloc(resource->value, sizeof (char*)*4);
			resource->num_values = 4;
			resource->value[3] = get_text(cur_node->children);	
		}
		if (strcmp(cur_node->name, "value4") == 0)
		{
			assert(FALSE);
		}
	}
}

