/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "workload_utils.h"
#include "fm_simulation_utils.h"

int index_of_next_event = 0;
LINKED_LIST *output_event_timeline;

/*---------------------------------------------------------------------------------------------
 * (function:  events_on_timeline)
 * returns if any events are on the timeline that matters.  Reoccurring events like timers
 * will be labelled unikely.
 *-------------------------------------------------------------------------------------------*/
short events_on_timeline(LINKED_LIST *timeline)
{
	LINKED_LIST_NODE *t;
        event_t *traverse;

	/* Go through the timeline */
        for (	t = LinkedListNodePeek(timeline);
		t != NULL;
		t = LinkedListNodeNext(t))
	{
		traverse = (event_t*)LinkedListNodeData(t);

		if ((traverse->type == INPUT_EVENT) || (traverse->type == COND_EVENT) || (traverse->type == OUTPUT_EVENT))
		{
			/* IF - there is an event that isn't just reoccuring */
			return TRUE;
		}
	}

	/* not used anymore so free it */
	LinkedListFree(timeline);
	return FALSE;
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_timeline)
 * Initilizes the timeline by putting all events in the input onto the list.  The assumption
 * is that the input workload is already sorted.
 *-------------------------------------------------------------------------------------------*/
void initialize_timeline(workload_t *w, LINKED_LIST *timeline)
{
	int i, j;

	for (i = 0; i < w->events->num_events; i++)
	{
		LinkedListInsert(timeline, (void*)w->events->events[i]);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_output)
 *-------------------------------------------------------------------------------------------*/
void initialize_output(workload_t *output)
{
	initialize_workload(output);

	output_event_timeline = LinkedListAlloc(ComparePointersBasedOnTime);
}

/*---------------------------------------------------------------------------------------------
 * (function: 	find_next_sim_step)
 * This goes through the timeline and checks if Outputs trigger Conditions.  The function 
 * continues to do this until we get to an input event that matters
 *-------------------------------------------------------------------------------------------*/
void find_next_sim_step(workload_t *w, workload_t *output, LINKED_LIST *timeline)
{
	LINKED_LIST_NODE *t;
        event_t *traverse;

	/* now figure out who is first, and if there are outputs before the next time step then make them happen */
	while ((t = LinkedListNext(timeline)) != NULL)
	{
       		traverse = (event_t*)t;

		if ((traverse->type == INPUT_EVENT) || (traverse->type == COND_EVENT) || (traverse->type == TIMING_EVENT))
		{
			/* IF - the next event should happen then return */
			LinkedListInsert(timeline, (void*)t);
			return;
		}
		else if (traverse->type == OUTPUT_EVENT)
		{
			/* IF the time is greater than or equal to an OUTPUT event then */
			/* when we get here, we've found a time in the output that could trigger a condition. */
			check_for_condition_trigger(w, traverse, timeline);

			/* now add to the output event list since we don't need to review it anymore */
			add_event_to_event_list(output, traverse->time, traverse->resources[0]->default_name, traverse->resources[0]->name, traverse->resources[0]->value, traverse->resources[0]->num_values, OUTPUT_EVENT, environment.minimum_operating_speed);
		}
	}

	/* get here if there are only ouputs that don't trigger conditions on the timeline */
	return;
}

/*---------------------------------------------------------------------------------------------
 * (function: check_for_condition_trigger)
 * Checks if an output triggers a condition by comparing the if part with the output details
 *-------------------------------------------------------------------------------------------*/
void check_for_condition_trigger(workload_t *w, event_t *output_event, LINKED_LIST *timeline)
{
	int i, j, k;
	short match = FALSE;

	for (i = 0; i < w->conditions->num_if_resource_pairs; i++)
	{
		match = FALSE;

		/* need to process next output time */
		for (j = 0; j < output_event->num_resources; j++)
		{
			char *resource_name;
			short to_free = FALSE;

			/* need to get theresource name for dfault names */
			if (output_event->resources[j]->name != NULL)
			{
				resource_name = output_event->resources[j]->name;
				to_free = FALSE;
			}
			else
			{
				resource_name = get_default_resource_name(output_event->resources[j]->default_name);
				to_free = TRUE;
			}

			if (strcmp(w->conditions->ifs[i]->name, resource_name) == 0)
			{
				/* IF - there is a match check if values match */
				match = TRUE;
				if (w->conditions->ifs[i]->num_values == output_event->resources[j]->num_values)
				{
					/* IF they have values then the number of values and value should also match */
					for (k = 0; k < w->conditions->ifs[i]->num_values; k++)
					{
						if (strcmp(w->conditions->ifs[i]->value[k], output_event->resources[j]->value[k]) != 0)
						{
							match = FALSE;
							break;
						}
					}
				}

				if (match == TRUE)
				{
					/* need to add to the condition_events list. */
					mpz_t *new_time;
					new_time = (mpz_t*)malloc(sizeof(mpz_t));

					mpz_init(*new_time);
					mpz_add(*new_time, *(w->conditions->then_times[i]), *(output_event->time));

					/* add the event to the list in its proper spot...can't happen before this output */
					add_event_to_timeline(timeline, new_time, w->conditions->then_resources[i]->name,  w->conditions->then_resources[i]->value, w->conditions->then_resources[i]->num_values, COND_EVENT); 
				}
			}

			if (to_free == TRUE)
			{
				/* free the resource name if it comes from a defualt name */
				free(resource_name);
			}
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: create_event_clump)
 * Goes into the timeline and grabs all events with the same time.  It creates a package
 * of all the resources occuring at the current time step.
 *-------------------------------------------------------------------------------------------*/
event_t *create_event_clump(LINKED_LIST *timeline)
{
	int j; 
	event_t *return_event;
	LINKED_LIST_NODE *t = NULL;
        event_t *traverse = NULL;

	/* setup the retrun structure */
	return_event = (event_t*)malloc(sizeof(event_t));
	return_event->num_resources = 0;
	return_event->resources = NULL;

	/* peek at the timeline if there is one */
	if (LinkedListSize(timeline) > 0)
	{
	        t = LinkedListNodePeek(timeline);
       		traverse = (event_t*)LinkedListNodeData(t);
		return_event->time = (mpz_t*)malloc(sizeof(mpz_t));
		mpz_init(*return_event->time);
		mpz_set(*return_event->time, *traverse->time); 
	}
	else
	{
		/* nothing in the schedule so just return */
		return return_event;
	}

       	while ((t = LinkedListNext(timeline)) != NULL)
	{
       		traverse = (event_t*)t;

		if ((mpz_cmp(*(return_event->time), *(traverse->time)) == 0) && (traverse->type != OUTPUT_EVENT))
		{	
			/* add it the resources to the list */
			for (j = 0; j < traverse->num_resources; j++)
			{	
				return_event->num_resources++;
				return_event->resources = (resource_t **)realloc(return_event->resources, sizeof(resource_t*)*return_event->num_resources);
				return_event->resources[return_event->num_resources-1] = traverse->resources[j];
			}

			/* free the event_t allocated in add to time line */
			if (traverse->type != INPUT_EVENT)
			{
				free(traverse->resources);
				free(traverse);
			}
		}
		else
		{
			/* Otherwise reinsert this element since we popped it */
			LinkedListInsert(timeline, (void*)t);
			break;
		}
	}

	return return_event;
}

/*---------------------------------------------------------------------------------------------
 * (function: free_event_clump)
 *-------------------------------------------------------------------------------------------*/
void free_event_clump(event_t *clump)
{
	/* setup the retrun structure */
	if (clump != NULL)
		free(clump);
}

/*---------------------------------------------------------------------------------------------
 * (function: add_event_to_timeline)
 *-------------------------------------------------------------------------------------------*/
void add_event_to_timeline(LINKED_LIST *list_of_events, mpz_t *time, char *resource_name, char **values, int num_values, short type)
{
	int i;
	event_t *to_add;

	to_add = (event_t *)malloc(sizeof(event_t));
	to_add->num_resources = 1;
	to_add->resources = (resource_t **)malloc(sizeof(resource_t*)*1);
	to_add->time = time;
	to_add->type = type;

	/* add a new resource */
	to_add->resources[0] = (resource_t *)malloc(sizeof(resource_t));

	/* store the name */
	to_add->resources[0]->name = resource_name; 

	/* store the values */
	to_add->resources[0]->num_values = num_values; 

	if (num_values > 0)
	{
		to_add->resources[0]->value = (char**)malloc(sizeof(char*)*num_values);
	}
	else
	{
		to_add->resources[0]->value = NULL;
	}

	for (i = 0; i < num_values; i++)
	{
		to_add->resources[0]->value[i] = values[i];
	}

	/* now add it to the linked list at the proper time.  This insert uses the ComparePointerBasedOnTime that has
	 * already been passed in at link list initialization. */
	LinkedListInsert(list_of_events, (void*)to_add);
}

/*---------------------------------------------------------------------------------------------
 * (function: ComparePointersBasedOnTime )         
 *      This function is used with any linked lists as the comparison component of a generic
 *      linked list.  This one is for checking the time.
 *      A return of -1 will cause the item to be inserted at this point.
 *-------------------------------------------------------------------------------------------*/                              
int ComparePointersBasedOnTime(foint a, foint b)
{
        event_t *new_entry = (event_t*) a; 
        event_t *element_in_list = (event_t*) b;

	if (mpz_cmp(*(new_entry->time), *(element_in_list->time)) > 0) // don't insert infront of element B if time A is not bigger than time B  
	   
	{
		/* return of 0 says to NOT insert before in the linked list */
		return 0;
	}
	else if ((mpz_cmp(*(new_entry->time), *(element_in_list->time)) == 0) && ((new_entry->type > OUTPUT_EVENT) && (element_in_list->type == OUTPUT_EVENT))) // do not insert A infront of B if B is an output and A is not an Output and the times are the same
	{
		/* return of 0 says to NOT insert before in the linked list */
		return 0;
	}
	return -1;          
}

/*---------------------------------------------------------------------------------------------
 * (function: debug_print_timeline)
 * Prints out the linked list of the timeline
 *-------------------------------------------------------------------------------------------*/
void debug_print_timeline(LINKED_LIST *timeline)
{
	LINKED_LIST_NODE *t;
        event_t *traverse;
	char *temp_string;

	if (LinkedListSize(timeline) == 0)
		return;

        for (	t = LinkedListNodePeek(timeline);
		t != NULL;
		t = LinkedListNodeNext(t))
	{
        	traverse = (event_t*)LinkedListNodeData(t);

		temp_string = mpz_get_str(NULL, 10, *(traverse->time));
		printf("Time: %s Name: %s ->", temp_string, traverse->resources[0]->name);
		free(temp_string);
	}
	printf("\n");
	fflush(stdout);
}
