/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "linked-list.h"
#include "aes.h"
#include "simulation_utils.h"

ae_device_state_t ae_state;
int recurring_events; // continuous timers

void initialize_aes_state(ae_device_state_t *state);
void ae_process_event_clump(LINKED_LIST *output_timeline, ae_device_state_t *state, event_t *clump_to_process);

/*---------------------------------------------------------------------------------------------
 * (function: golden_functional_model_for_aes)
 *-------------------------------------------------------------------------------------------*/
void golden_functional_model_for_aes(workload_t *workload, char *output_file)
{
	int total_number_of_events = workload->events.num_events;
	event_t *next_clump;

	/* number of self inserting events ... in aes logic it's the time sequence */
	recurring_events = 1;

	/* initializes the output data structure */
	initialize_output(&output);
	/* initializes the local state of the functional design */
	initialize_aes_state(&aes_state);
	/* initlize the timeline */
	initialize_timeline(workload, output_event_timeline);

	/* event processing loop */
	while (events_on_timeline(output_event_timeline) == TRUE)
	{
		/* build a list of all events that should happen next */
		next_clump = create_event_clump(output_event_timeline); 

		/* process events against chip behaviour and conditions */		
		ae_process_event_clump(output_event_timeline, &aes_state, next_clump);

		/* find next sim time by removing uneeded output events and trigerring conditions */
		find_next_sim_step(workload, &output, output_event_timeline);

		/* free the clump */	
		free(next_clump->resources);
		free(next_clump);
	}

	/* now that all the events are processed, create the output file */
	generate_the_output_workload_file(output_file, &output, workload);

	/* cleanup */
}

/*---------------------------------------------------------------------------------------------
 * (function: initialize_aes_state)
 *-------------------------------------------------------------------------------------------*/
void initialize_aes_state(ae_device_state_t *state)
{
}

/*---------------------------------------------------------------------------------------------
 * (function: ae_process_event_clump)
 *-------------------------------------------------------------------------------------------*/
void ae_process_event_clump(LINKED_LIST *output_timeline, ae_device_state_t *state, event_t *clump_to_process)
{
	int i, index_of_resource; 
	char *resource_name;
	mpz_t *output_time = (mpz_t*)malloc(sizeof(mpz_t));
	short output_time_used = FALSE;

	/* setup the output time delay ... this probably should have details of all the signals in later versions */
	mpz_init(*output_time);
	mpz_set(*output_time, *(clump_to_process->time));
	mpz_add_ui(*output_time, *output_time, (unsigned int)13000); // add 13us ... this is a hard limit on Y

	/* now go through all the events in the current clump which represents a timestep */
	for (index_of_resource = 0; index_of_resource < clump_to_process->num_resources; index_of_resource++)
	{	
		char *name = clump_to_process->resources[index_of_resource]->name;
		int num_values = clump_to_process->resources[index_of_resource]->num_values;
		char **value = clump_to_process->resources[index_of_resource]->value;
		
		/* ---------------------------------------------------------------------------------------------------------------
		 * This is the unique part for handling each resource type
		 * -------------------------------------------------------------------------------------------------------------*/
		if (strcmp(name, "reset") == 0)
		{
			if (strcmp(value[0], "1") == 0)
			{
			}
			else if (strcmp(value[0], "0") == 0)
			{
			}
		}
		else if (strcmp(name, "other_resource") == 0)
		{
		}
	}

	if (output_time_used == FALSE)
	{
		/* free the time if it wasn't used ... can happen since some inputs only change internal data and not external outputs */
		mpz_clear(*output_time);
		free(output_time);
	}
}
