/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "fm_read_xml_input.h"
#include "read_xml_config_file.h"
#include "read_xml_default_waveforms.h"
#include "workload_utils.h"
#include "fm_aes.h"
#include "fm_glue_logic.h"
#include "fm_lziv.h"
#include "fm_portkey.h"
#include "fm_2dconv.h"
#include "fm_bridge.h"
#include "wave_generation.h"

void do_functional_model_simulation();
void do_workload_generation();

int main(int argc, char **argv)
{
	if (argc != 4)
	{
		printf("usage: benchamrk_infra.exe <config_file> <default_workload_file> <environment_file>\n");
		exit(-1);
	}

	/* read the config file to figure out what to do */
	read_config_file(argv[1]);
	/* read the default waveforms so we have the parameters to generate the needed waveforms */
	read_default_waveforms(argv[2]);
	/* read the envirnment specification file */
	read_environment(argv[3]);

	/* generate workloads, and output these workloads to a file format -- either verilog state machine or xml */
	do_workload_generation();

	/* do the golden functional model testing for workloads specified in the config file */
	do_functional_model_simulation();
} 

/*---------------------------------------------------------------------------------------------
 * (function: do_functional_model_simulation)
 * This is the high-level control for functional model simulation
 *
 * Note that functional simulation does workload generation and simulation, so doesn't need
 * workload generation.
 *-------------------------------------------------------------------------------------------*/
void do_functional_model_simulation()
{
	int i, j;
	void (*golden_function)(workload_t*, char*);
	int benchmark_number;
	char file_name[4096];
	workload_t *current_workload;

	if (configuration.gf != NULL)
	{
		for (i = 0; i < configuration.gf->num_benchmarks; i++)
		{
			/* output XML if that is the output format */
			if (configuration.gf->output_type == XML)
			{
				/* setup the appropriate functions and naming for the given benchmark and
				 * the simulation function to use */
				if (strcmp(configuration.gf->benchmarks[i]->design_name, "portkey") == 0)
				{
					benchmark_number = 1;
					golden_function = &golden_functional_model_for_keypad;
				}
				else if (strcmp(configuration.gf->benchmarks[i]->design_name, "glue") == 0)
				{
					benchmark_number = 2;
					golden_function = &golden_functional_model_for_glue_logic;
				}
				else if (strcmp(configuration.gf->benchmarks[i]->design_name, "aes") == 0)
				{
					benchmark_number = 3;
					golden_function = &golden_functional_model_for_aes;
				}
				else if (strcmp(configuration.gf->benchmarks[i]->design_name, "lziv") == 0)
				{
					benchmark_number = 4;
					golden_function = &golden_functional_model_for_lziv;
				}
				else if (strcmp(configuration.gf->benchmarks[i]->design_name, "bridge") == 0)
				{
					benchmark_number = 5;
					golden_function = &golden_functional_model_for_bridge;
				}
				else if (strcmp(configuration.gf->benchmarks[i]->design_name, "2dconv") == 0)
				{
					benchmark_number = 6;
					golden_function = &golden_functional_model_for_2dconv;
				}
	
				/* this does all the explicit file names included in the config file, and these
				 * will be input workload xml files - we expect that this will be rarely used since
				 * the current code can generate the workloads. */
				for (j = 0; j < configuration.gf->benchmarks[i]->num_workload_file_names; j++)
				{
					/* we append out, so currently we don't handle file paths */
					sprintf(file_name, "out_%s", configuration.gf->benchmarks[i]->workload_file_names[j]); 
	
					/* read in the xml workload details */
					current_workload = read_workload_input(configuration.gf->benchmarks[i]->workload_file_names[j]);
	
					/* call the simulation function */
					(*golden_function)(current_workload, file_name);
				}

				/* This is the non input XML file where we just go through the default waveform and
				 * markov percentages to generate the workloads */
				for (j = 0; j < configuration.gf->benchmarks[i]->num_workloads; j++)
				{
					sprintf(file_name, "out_w.%d.GH09.B.%d.%s.xml", configuration.gf->benchmarks[i]->workloads[j], benchmark_number, configuration.gf->benchmarks[i]->design_name); 
	
					/* find the workload in the default wave list */
					current_workload = generate_wave(configuration.gf->benchmarks[i]->design_name, configuration.gf->benchmarks[i]->workloads[j]);
	
					/* call the simulation function */
					(*golden_function)(current_workload, file_name);

					free_workload(current_workload);
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: do_workload_generation)
 * This function creates workloads.  These workloads are created in a datastructure and then
 * it can be passed out in an output format.
 *
 * Current formats:
 * 	- XML
 * 	- Internal data structure
 *
 * Note that the simulation functionality does the generation too.
 *-------------------------------------------------------------------------------------------*/
void do_workload_generation()
{
	int i, j;
	char file_name[4096];
	int benchmark_number;
	workload_t *current_workload;

	if (configuration.wg != NULL)
	{
		for (i = 0; i < configuration.wg->num_benchmarks; i++)
		{
			/* go through each workload and generate the pulse wave */
			for (j = 0; j < configuration.wg->benchmarks[i]->num_workloads; j++)
			{
				/* output XML if that is the output format */
				if (configuration.wg->output_type == XML)
				{
					/* setup the appropriate functions and naming for the given benchmark */
					if (strcmp(configuration.wg->benchmarks[i]->design_name, "portkey") == 0)
					{
						benchmark_number = 1;
					}
					else if (strcmp(configuration.wg->benchmarks[i]->design_name, "glue") == 0)
					{
						benchmark_number = 2;
					}
					else if (strcmp(configuration.wg->benchmarks[i]->design_name, "aes") == 0)
					{
						benchmark_number = 3;
					}
					else if (strcmp(configuration.wg->benchmarks[i]->design_name, "lziv") == 0)
					{
						benchmark_number = 4;
					}
					else if (strcmp(configuration.wg->benchmarks[i]->design_name, "bridge") == 0)
					{
						benchmark_number = 5;
					}
					else if (strcmp(configuration.wg->benchmarks[i]->design_name, "2dconv") == 0)
					{
						benchmark_number = 6;
					}

					sprintf(file_name, "w.%d.GH09.B.%d.%s.xml", configuration.wg->benchmarks[i]->workloads[j], benchmark_number, configuration.wg->benchmarks[i]->design_name); 

					/* generate the internal data structure for the specified wave based on benchmark number and workload number */
					current_workload = generate_wave(configuration.wg->benchmarks[i]->design_name, configuration.wg->benchmarks[i]->workloads[j]);

					/* create the output file */
					generate_an_output_workload_file(file_name, current_workload, current_workload->design_name, configuration.wg->benchmarks[i]->workloads[j]);

					free_workload(current_workload);
				}
				if (configuration.wg->output_type == VERILOG)
				{
					/* IF - this is a verilog file then we need to create the verilog code to to do this */
					assert(FALSE);
				}
			}
		}
	}
}
