/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "types.h"
#include "globals.h"
#include "util.h"
#include "gmp.h"
#include "read_xml_config_file.h"

config_t configuration;

void read_top_con(xmlNode * a_node);
void read_functional_model(xmlNode * a_node, config_t *config);
void read_workload_generation(xmlNode *a_node, config_t *config);
void read_workload_descriptions(xmlNode * a_node, workload_description_t **benchmarks, int index);

char* get_specified_tag_text(xmlNode * a_node, char *specified_tag);
char* get_text(xmlNode * a_node);

/*---------------------------------------------------------------------------------------------
 * (function: read_config_file)
 * This reads an XML config file that specifies what we will do in the tool.
 *
 * See config_t in types.h to see the data structures used in this read.
 *-------------------------------------------------------------------------------------------*/
void read_config_file(char *file_name)
{
	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;

	/* this initialize the library and check potential ABI mismatches between the version it was compiled for and the actual shared library used. */
	LIBXML_TEST_VERSION

	/*parse the file and get the DOM */
	doc = xmlParseFile(file_name);

	if (doc == NULL) 
	{
		printf("error: could not parse file LIB file\n");
	}

	/*Get the root element node */
	root_element = xmlDocGetRootElement(doc);

	/* parse in the XML to data structs */
	read_top_con(root_element);

	/*free the document */
	xmlFreeDoc(doc);

	/* Free the global variables that may have been allocated by the parser.  */
	xmlCleanupParser();
}


/*---------------------------------------------------------------------------------------------
 * (function: read_top_con)
 *-------------------------------------------------------------------------------------------*/
void read_top_con(xmlNode * a_node)
{
	int i, j;
	xmlNode *top_node = NULL;
	xmlNode *cur_node = NULL;
	xmlChar *number_tfe_string = NULL;

	for (top_node = a_node; top_node; top_node = top_node->next) 
	{
		if (strcmp(top_node->name, "config") == 0)
		{
			/* IF - Make sure this is the top node */
			configuration.wg = NULL;
			configuration.gf = NULL;

			for (cur_node = top_node->children; cur_node; cur_node = cur_node->next) 
			{
				if (cur_node->type == XML_ELEMENT_NODE) 
				{
					/* IF - these top level nodes should be lements */
					if (strcmp(cur_node->name, "workload_generation") == 0)
					{
						read_workload_generation(cur_node->children, &configuration);
					}
					else if (strcmp(cur_node->name, "golden_functional_model") == 0)
					{
						read_functional_model(cur_node->children, &configuration);
					}
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_functional_model)
 *-------------------------------------------------------------------------------------------*/
void read_functional_model(xmlNode * a_node, config_t *config)
{
	xmlNode *cur_node = NULL;
	xmlNode *cur_node2 = NULL;
	golden_func_t *gf;

	config->gf = (golden_func_t *)malloc(sizeof(golden_func_t));
	gf = config->gf;
	gf->output_type = -1;
	gf->num_benchmarks = 0;
	gf->benchmarks = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "benchmarks") == 0)
		{
			for (cur_node2 = cur_node->children; cur_node2; cur_node2 = cur_node2->next) 
			{
				if (strcmp(cur_node2->name, "benchmark") == 0)
				{
					gf->benchmarks = (workload_description_t**)realloc(gf->benchmarks, sizeof(workload_description_t*)*(gf->num_benchmarks+1));
					gf->num_benchmarks++;

					read_workload_descriptions(cur_node2->children, gf->benchmarks, gf->num_benchmarks-1);
				}
			}


		}
		else if (strcmp(cur_node->name, "output_type") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			if (strcmp(temp_string, "xml") == 0)
			{
				gf->output_type = XML;
			}
			free(temp_string);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_workload_generation)
 *-------------------------------------------------------------------------------------------*/
void read_workload_generation(xmlNode *a_node, config_t *config)
{
	xmlNode *cur_node = NULL;
	xmlNode *cur_node2 = NULL;
	workload_gen_t *wg;

	config->wg = (workload_gen_t *)malloc(sizeof(workload_gen_t));
	wg = config->wg;
	wg->output_type = -1;
	wg->num_benchmarks = 0;
	wg->benchmarks = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "benchmarks") == 0)
		{
			for (cur_node2 = cur_node->children; cur_node2; cur_node2 = cur_node2->next) 
			{
				if (strcmp(cur_node2->name, "benchmark") == 0)
				{
					wg->benchmarks = (workload_description_t**)realloc(wg->benchmarks, sizeof(workload_description_t*)*(wg->num_benchmarks+1));
					wg->num_benchmarks++;

					read_workload_descriptions(cur_node2->children, wg->benchmarks, wg->num_benchmarks-1);
				}
			}

		}
		else if (strcmp(cur_node->name, "output_type") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			if (strcmp(temp_string, "xml") == 0)
			{
				wg->output_type = XML;
			}
			else if (strcmp(temp_string, "verilog") == 0)
			{
				wg->output_type = VERILOG;
			}
			free(temp_string);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_workload_descriptions)
 *-------------------------------------------------------------------------------------------*/
void read_workload_descriptions(xmlNode * a_node, workload_description_t **benchmarks, int index)
{
	xmlNode *cur_node = NULL;

	benchmarks[index] = (workload_description_t *)malloc(sizeof(workload_description_t));
	benchmarks[index]->design_name = NULL;
	benchmarks[index]->workloads = NULL;
	benchmarks[index]->num_workloads = 0;
	benchmarks[index]->num_workload_file_names = 0;
	benchmarks[index]->workload_file_names = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "name") == 0)
		{
			benchmarks[index]->design_name = get_text(cur_node->children);
		}
		else if (strcmp(cur_node->name, "workload") == 0)
		{
			benchmarks[index]->num_workloads++;
			benchmarks[index]->workloads = (int*)realloc(benchmarks[index]->workloads, sizeof(int)*benchmarks[index]->num_workloads);

			char *temp_string =  get_text(cur_node->children);
			benchmarks[index]->workloads[benchmarks[index]->num_workloads-1] = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "workload_file_name") == 0)
		{
			benchmarks[index]->workload_file_names = (char**)realloc(benchmarks[index]->workload_file_names, sizeof(char*)*(benchmarks[index]->num_workload_file_names+1));
			benchmarks[index]->num_workload_file_names++;

			benchmarks[index]->workload_file_names[benchmarks[index]->num_workload_file_names-1] = get_text(cur_node->children);
		}
	}
}
