/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "types.h"
#include "globals.h"
#include "util.h"
#include "gmp.h"
#include "read_xml_config_file.h"

waves_t default_workloads;

void read_top_wave(xmlNode * a_node);
void read_waves(xmlNode * a_node, waves_t *waves);
void read_waveform_params(xmlNode * a_node, workload_waves_t **wave, int index);

/*---------------------------------------------------------------------------------------------
 * (function: read_default_waveforms)
 * Reads the waveform file that describes the parameters of the wave of events.  See wave_t
 * for a description of the data structure.
 *-------------------------------------------------------------------------------------------*/
void read_default_waveforms(char *file_name)
{
	xmlDoc *doc = NULL;
	xmlNode *root_element = NULL;

	/* this initialize the library and check potential ABI mismatches between the version it was compiled for and the actual shared library used. */
	LIBXML_TEST_VERSION

	/*parse the file and get the DOM */
	doc = xmlParseFile(file_name);

	if (doc == NULL) 
	{
		printf("error: could not parse file LIB file\n");
	}

	/*Get the root element node */
	root_element = xmlDocGetRootElement(doc);

	/* parse in the XML to data structs */
	read_top_wave(root_element);

	/*free the document */
	xmlFreeDoc(doc);

	/* Free the global variables that may have been allocated by the parser.  */
	xmlCleanupParser();
}

/*---------------------------------------------------------------------------------------------
 * (function: read_top_wave)
 *-------------------------------------------------------------------------------------------*/
void read_top_wave(xmlNode * a_node)
{
	int i, j;
	xmlNode *top_node = NULL;
	xmlChar *number_tfe_string = NULL;

	for (top_node = a_node; top_node; top_node = top_node->next) 
	{
		if (strcmp(top_node->name, "waves") == 0)
		{
			read_waves(top_node->children, &default_workloads);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_waves)
 *-------------------------------------------------------------------------------------------*/
void read_waves(xmlNode * a_node, waves_t *waves)
{
	xmlNode *cur_node = NULL;

	waves->num_workload_waves = 0;
	waves->workload_waves = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "workload_params") == 0)
		{
			waves->workload_waves = (workload_waves_t**)realloc(waves->workload_waves, sizeof(workload_waves_t*)*(waves->num_workload_waves+1));
			waves->num_workload_waves++;

			read_waveform_params(cur_node->children, waves->workload_waves, waves->num_workload_waves-1);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: read_waveform_params)
 *-------------------------------------------------------------------------------------------*/
void read_waveform_params(xmlNode * a_node, workload_waves_t **wave, int index)
{
	xmlNode *cur_node = NULL;

	wave[index] = (workload_waves_t *)malloc(sizeof(workload_waves_t));
	wave[index]->name = NULL;
	wave[index]->design_name = NULL;
	wave[index]->workload_value = NULL;
	wave[index]->num_workload_value = 0;
	wave[index]->workload_string = NULL;
	wave[index]->num_workload_string = 0;
	wave[index]->seed = 1; // default value

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, "name") == 0)
		{
			wave[index]->name = get_text(cur_node->children);
		}
		else if (strcmp(cur_node->name, "design_name") == 0)
		{
			wave[index]->design_name = get_text(cur_node->children);
		}
		else if (strcmp(cur_node->name, "total_time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->total_time = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "X_time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->X_ns = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "Y_time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->Y_ns = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "Z_time") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->Z_ns = time_to_nanoseconds(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "X_deviation") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->X_deviation = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "Y_deviation") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->Y_deviation = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "Z_deviation") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->Z_deviation = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "seed") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->seed = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "workload_value") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->num_workload_value ++;
			wave[index]->workload_value = (int*)realloc(wave[index]->workload_value, sizeof(int)*wave[index]->num_workload_value);
			wave[index]->workload_value[wave[index]->num_workload_value-1] = atoi(temp_string);
			free(temp_string);
		}
		else if (strcmp(cur_node->name, "workload_string") == 0)
		{
			char *temp_string = get_text(cur_node->children);
			wave[index]->num_workload_string ++;
			wave[index]->workload_string = (char**)realloc(wave[index]->workload_string, sizeof(char*)*wave[index]->num_workload_string);
			wave[index]->workload_string[wave[index]->num_workload_string-1] = temp_string;
		}
	}
}
