/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "types.h"
#include "globals.h"
#include "bitvector.h"

unsigned long s1, s2, s3, b; // s1, s2, s3 must be initialized

/*---------------------------------------------------------------------------------------------
 * (function: my_rand)
 *
 * Calls the random function generator that we use
 *-------------------------------------------------------------------------------------------*/
unsigned long my_rand()
{
	return (int)(taus88_pseudo_random_number_generator());
	// old return was for when the random nuber returned a double between 0 and 1
	//return (int)(taus88_pseudo_random_number_generator() * 32000);
}

/*---------------------------------------------------------------------------------------------
 * (function: taus88_seed)
 * seed the random number generator
 *-------------------------------------------------------------------------------------------*/
void taus88_seed(unsigned long seed)
{
	s1 = seed+2147483645;
	s2 = seed+2147483646;
	s3 = seed+2147483647;
}

/*---------------------------------------------------------------------------------------------
 * (function: taus88_pseudo_random_number_generator)
 * Generates a value between 0 and 1.  Period = 2^113
 * Taken from the paper: Maximally Equidistributed Combined Tausworthe Generators
 *-------------------------------------------------------------------------------------------*/
unsigned long taus88_pseudo_random_number_generator()
{
	b = (((s1 << 13) ^ s1) >> 19);
	s1 = (((s1 & 4294967294UL) << 12) ^ b); // the UL is to say unsigned long for the compiler
	b = (((s2 << 2) ^ s2) >> 25);
	s2 = (((s2 & 4294967288UL) << 4) ^ b);
	b = (((s3 << 3) ^ s3) >> 11);
	s3 = (((s3 & 4294967280UL) >> 17) ^ b);
	return ((s1 ^ s2 ^ s3));// * 2.3283064365e-10);
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_bitvector_to_hex_string)
 * assume index 0 -> 00000011 <- index 7
 *                 0x0   3 
 *-------------------------------------------------------------------------------------------*/
char *convert_bitvector_to_hex_string(bit_vector *vector)
{
	int index = 0;
	int i;
	char *return_string;
	short mask = 1;
	char temp_string[3];
	int value;

	assert(vector->num_bits % 4 == 0);

	return_string = (char*)malloc(sizeof(char)*(vector->num_bits/4+1));

	while (index*4 < vector->num_bits)
	{
		mask = 1 << 3;
		//mask = 1;
		value = 0;
		for (i = 0; i < 4; i++)
		{
			if (get_bit(vector, index*4+i) == TRUE)
			{
				value += mask;
			}
			mask = mask >> 1;
			//mask = mask << 1;
		}

		sprintf(temp_string, "%x", value);
		return_string[index] = temp_string[0];

		index ++;
	}

	return_string[vector->num_bits/4] = '\0';

	return return_string;
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_hex_string_of_size_to_int)
 *-------------------------------------------------------------------------------------------*/
int convert_hex_string_of_size_to_int(char *orig_string, int size)
{
	int i;
	int return_value = 0;
	char temp[2];
	int index = size-1;

	for (i = 0; i < size; i++)
	{
		if (isdigit(orig_string[i]))
		{
			sprintf(temp, "%c", orig_string[i]);
			return_value += (int)(atoi(temp) * my_power(16, index));
		}
		else if ((orig_string[i] == 'a') || (orig_string[i] == 'A'))
		{
			return_value += (int)(10 * my_power(16, index));
		}
		else if ((orig_string[i] == 'b') || (orig_string[i] == 'B'))
		{
			return_value += (int)(11 * my_power(16, index));
		}
		else if ((orig_string[i] == 'c') || (orig_string[i] == 'C'))
		{
			return_value += (int)(12 * my_power(16, index));
		}
		else if ((orig_string[i] == 'd') || (orig_string[i] == 'D'))
		{
			return_value += (int)(13 * my_power(16, index));
		}
		else if ((orig_string[i] == 'e') || (orig_string[i] == 'E'))
		{
			return_value += (int)(14 * my_power(16, index));
		}
		else if ((orig_string[i] == 'f') || (orig_string[i] == 'F'))
		{
			return_value += (int)(15 * my_power(16, index));
		}
		index --;
	}	
	
	return return_value;
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_hex_string_to_int)
 *-------------------------------------------------------------------------------------------*/
int convert_hex_string_to_int(char *orig_string)
{
	int i;
	int return_value = 0;
	char temp[2];
	int index = 7;

	for (i = 0; i < 8; i++)
	{
		if (isdigit(orig_string[i]))
		{
			sprintf(temp, "%c", orig_string[i]);
			return_value += (int)(atoi(temp) * my_power(16, index));
		}
		else if ((orig_string[i] == 'a') || (orig_string[i] == 'A'))
		{
			return_value += (int)(10 * my_power(16, index));
		}
		else if ((orig_string[i] == 'b') || (orig_string[i] == 'B'))
		{
			return_value += (int)(11 * my_power(16, index));
		}
		else if ((orig_string[i] == 'c') || (orig_string[i] == 'C'))
		{
			return_value += (int)(12 * my_power(16, index));
		}
		else if ((orig_string[i] == 'd') || (orig_string[i] == 'D'))
		{
			return_value += (int)(13 * my_power(16, index));
		}
		else if ((orig_string[i] == 'e') || (orig_string[i] == 'E'))
		{
			return_value += (int)(14 * my_power(16, index));
		}
		else if ((orig_string[i] == 'f') || (orig_string[i] == 'F'))
		{
			return_value += (int)(15 * my_power(16, index));
		}
		index --;
	}	
	
	return return_value;
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_hex_string_to_short)
 *-------------------------------------------------------------------------------------------*/
short convert_hex_string_to_short(char *orig_string, int num_chars)
{
	int i;
	short return_value = 0;
	char temp[2];
	int index = num_chars - 1;

	assert(num_chars <= 2);

	for (i = 0; i < num_chars; i++)
	{
		if (isdigit(orig_string[i]))
		{
			sprintf(temp, "%c", orig_string[i]);
			return_value += (short)(atoi(temp) * my_power(16, index));
		}
		else if ((orig_string[i] == 'a') || (orig_string[i] == 'A'))
		{
			return_value += (short)(10 * my_power(16, index));
		}
		else if ((orig_string[i] == 'b') || (orig_string[i] == 'B'))
		{
			return_value += (short)(11 * my_power(16, index));
		}
		else if ((orig_string[i] == 'c') || (orig_string[i] == 'C'))
		{
			return_value += (short)(12 * my_power(16, index));
		}
		else if ((orig_string[i] == 'd') || (orig_string[i] == 'D'))
		{
			return_value += (short)(13 * my_power(16, index));
		}
		else if ((orig_string[i] == 'e') || (orig_string[i] == 'E'))
		{
			return_value += (short)(14 * my_power(16, index));
		}
		else if ((orig_string[i] == 'f') || (orig_string[i] == 'F'))
		{
			return_value += (short)(15 * my_power(16, index));
		}
		index --;
	}	
	
	return return_value;
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_bit_string_lsb_to_int)
 * Assumes character string is lsb to msb.  For example binary string 110 converts to 3
 *-------------------------------------------------------------------------------------------*/
int convert_bit_string_lsb_to_int(char *orig_string, int num_chars)
{
	int i;
	int return_val = 0;
	int mask = 1;

	for (i = 0; i < num_chars; i++)
	{
		if (orig_string[i] == '1')
		{
			return_val += mask;
		}
		mask = mask << 1;
	}	
	
	return return_val;
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_bit_string_msb_to_int)
 * Assumes character string is msb to lsb.  For example binary string 011 converts to 3
 *-------------------------------------------------------------------------------------------*/
int convert_bit_string_msb_to_int(char *orig_string, int num_chars)
{
	int i;
	int return_val = 0;
	int mask = 1;

	for (i = num_chars-1; i >= 0; i--)
	{
		if (orig_string[i] == '1')
		{
			return_val += mask;
		}
		mask = mask << 1;
	}	
	
	return return_val;
}
/*---------------------------------------------------------------------------------------------
 * (function: convert_int_to_bit_string)
 * Outpus a string lsb to msb.  For example, 3 becomes "110"
 *-------------------------------------------------------------------------------------------*/
char *convert_int_to_bit_string(int orig_int, int num_bits)
{
	int i;
	char *return_val = (char*)malloc(sizeof(char)*(num_bits+1));
	int mask = 1;

	for (i = 0; i < num_bits; i++)
	{
		if((mask & orig_int) > 0)
		{
			return_val[i] = '1';	
		}
		else
		{
			return_val[i] = '0';	
		}
		mask = mask << 1;
	}
	return_val[num_bits] = '\0';
	
	return return_val;
}

/*---------------------------------------------------------------------------------------------
 * (function: time_to_nanoseconds)
 * Converts a character string in the form 10min or 10s or 1000ms etc. to a time in nanoseconds
 *-------------------------------------------------------------------------------------------*/
mpz_t *time_to_nanoseconds(char *time)
{
	mpz_t *nanoseconds;
	mpz_t conversion;

	int i;
	int number_of_digits = 0;
	int where_unit_starts = 0;
	int number_of_unit_chars = 0;
	int space = 0;
	char *just_digits;
	char *suffix;

	nanoseconds = (mpz_t*)malloc(sizeof(mpz_t));

	/* analyze the string in the form "[0-9].[a-zA-z].".  For example "1000us" or "1000 us" */ 
	for (i= 0; i < strlen(time); i++)
	{
		if (isdigit(time[i]))
		{
			number_of_digits ++;
		}
		else if (isalpha(time[i]))
		{
			if (where_unit_starts == 0)
			{
				where_unit_starts = i;
			}
			number_of_unit_chars ++;
		}
		else if (isspace(time[i]))
		{
			space++;
		}
		else
		{
			assert(FALSE);
		}
	}

	assert(space + number_of_digits + number_of_unit_chars == strlen(time));

	/* convert the just digits into a big number */
	just_digits = (char*)malloc(sizeof(char)*(number_of_digits+1));
	strncpy(just_digits, time, number_of_digits);
	just_digits[number_of_digits] = '\0';
	mpz_init_set_str(*nanoseconds, just_digits, 10);

	/* convert the suffix and then multiply the value by the nano second */
	suffix = (char*)malloc(sizeof(char)*(number_of_unit_chars+1));
	strncpy(suffix, &(time[where_unit_starts]), number_of_unit_chars);
	suffix[number_of_unit_chars] = '\0';
	if (strcmp(suffix, "min") == 0)
	{
		mpz_init_set_str(conversion, "60000000000", 10);
	}
	else if (strcmp(suffix, "s") == 0)
	{
		mpz_init_set_str(conversion, "1000000000", 10);
	}
	else if (strcmp(suffix, "ms") == 0)
	{
		mpz_init_set_str(conversion, "1000000", 10);
	}
	else if (strcmp(suffix, "us") == 0)
	{
		mpz_init_set_str(conversion, "1000", 10);
	}
	else if (strcmp(suffix, "ns") == 0)
	{
		mpz_init_set_str(conversion, "1", 10);
	}
	else 
	{
		assert(FALSE);
	}
	mpz_mul(*nanoseconds, *nanoseconds, conversion);

	/* free allocated memory */
	free(just_digits);
	free(suffix);
	mpz_clear(conversion);

	return nanoseconds;
}

/*---------------------------------------------------------------------------------------------
 * (function: bit_to_string)
 * Assumes the string is already allocated
 *-------------------------------------------------------------------------------------------*/
void bit_to_string(char *bit_string, short bit)
{
	if (bit == ONE)
	{
		sprintf(bit_string, "1");
	}
	else if (bit == ZERO)
	{
		sprintf(bit_string, "0");
	}
	else if (bit == bitX)
	{
		sprintf(bit_string, "X");
	}
	else if (bit == bitZ)
	{
		sprintf(bit_string, "Z");
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: convert_bit_string_lsb_to_int)
 * Prints out the time
 *-------------------------------------------------------------------------------------------*/
void debug_time (mpz_t *time)
{
	char *time_string;
	time_string = mpz_get_str(NULL, 10, *(time));
	printf("time %s \n", time_string);
	free(time_string);
}

/*---------------------------------------------------------------------------------------------
 * (function: my_power)
 *      My own simple power function
 *-------------------------------------------------------------------------------------------*/
long long int my_power(long long int x, long long int y)
{
	int i;
	long long int value;

	if (y == 0)
	{
		return 1;
	}

	value = x;

	for (i = 1; i < y; i++)
	{
		value *= x;
	}

	return value;
}

/*---------------------------------------------------------------------------------------------
 * (function: get_specified_tag_text)
 * 	This function grabs XML text under the knowledge that there should be text.
 *-------------------------------------------------------------------------------------------*/
char* get_specified_tag_text(xmlNode * a_node, char *specified_tag)
{
	xmlNode *cur_node = NULL;
	char *return_string = NULL;

	/* get the name of the device */	
	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		if (strcmp(cur_node->name, specified_tag) == 0)
		{
			return_string = get_text(cur_node->children);
		}
	}
	assert(return_string != NULL);

	return return_string;
}

/*---------------------------------------------------------------------------------------------
 * (function: get_text)
 * 	This function grabs XML text under the knowledge that there should be text.
 *-------------------------------------------------------------------------------------------*/
char* get_text(xmlNode * a_node)
{
	xmlNode *cur_node = NULL;
	xmlChar *return_word = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
	{
		assert(return_word == NULL);
		assert((cur_node->type == XML_TEXT_NODE)&&(xmlIsBlankNode(cur_node) == FALSE));

		/* Get the word */
		return_word = xmlNodeGetContent(cur_node);
	}

	return (return_word);
}


/*---------------------------------------------------------------------------------------------
 * (function: lookup_default_waveform)
 *-------------------------------------------------------------------------------------------*/
workload_waves_t *lookup_default_waveform(char *lookup_string)
{
	int i;
		
	for (i = 0; i < default_workloads.num_workload_waves; i++)
	{
		if (strcmp(lookup_string, default_workloads.workload_waves[i]->name) == 0)
		{
			/* if we find it then */
			return default_workloads.workload_waves[i];
		}
	}

	return NULL;
}

/*---------------------------------------------------------------------------------------------
 * (function: create_32_hex_string)
 * creates a 32 bit random hex string
 *-------------------------------------------------------------------------------------------*/
char *create_32_hex_string()
{
	int i;
	char *return_string;
	char string_choices[17] = "0123456789abcdef";
	int rand_char;

	return_string = (char*)malloc(sizeof(char)*33);

	for (i = 0; i < 32; i++)
	{
		return_string[i] = string_choices[my_rand() % 16];
		
	}

	return_string[32] = '\0';

	return return_string;
}

/*---------------------------------------------------------------------------------------------
 * (function: create_hex_string)
 * creates a random hex string of defined size
 *-------------------------------------------------------------------------------------------*/
char *create_hex_string(int block_size)
{
	int i;
	char *return_string;
	char string_choices[17] = "0123456789abcdef";
	//char string_choices[17] = "0123400000000000";
	int rand_char;

	return_string = (char*)malloc(sizeof(char)*(block_size+1));

	for (i = 0; i < block_size; i++)
	{
		return_string[i] = string_choices[my_rand() % 16];
	}

	return_string[block_size] = '\0';

	return return_string;
}
