/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "workload_utils.h"
#include "wave_generation.h"
#include "wg_portkey.h"
#include "wg_aes.h"
#include "wg_lziv.h"
#include "wg_glue_logic.h"
#include "wg_bridge.h"
#include "wg_2dconv.h"

/*---------------------------------------------------------------------------------------------
 * (function: generate_wave_generation)
 * Generate waves based on the design name and the workload number
 *-------------------------------------------------------------------------------------------*/
workload_t *generate_wave(char *design_name, int workload_number)
{
	int i;
	void *event_function;
	void (*header_function)(workload_t*, workload_waves_t*);

	short bm_type = 0;
	workload_t *current_workload;
	int index_found = -1;

	/* create the workload structure that represents the input wave of events */
	current_workload = (workload_t*)malloc(sizeof(workload_t));

	/* find the index of waves with the same name as the design name  ... this is read in the default waveform file */
	for (i = 0; i < default_workloads.num_workload_waves; i++)
	{
		if ((strcmp(default_workloads.workload_waves[i]->design_name, design_name) == 0) && (atoi(default_workloads.workload_waves[i]->name) == workload_number))
		{
			index_found = i;
			break;
		}
	}

	if (index_found != -1)
	{
		/* initialize the seed ... we try to make constant so we can regenerate in hardware form if we want to ... plus this is a good random number generator */
		taus88_seed(default_workloads.workload_waves[i]->seed);
	
		/* setup the appropriate functions and naming for the given benchmark */
		if (strcmp(design_name, "portkey") == 0)
		{
			event_function = &do_next_portkey_event;
			header_function = &do_portkey_top;
		}
		else if (strcmp(design_name, "glue") == 0)
		{
			event_function = &do_next_glue_logic_input;
			header_function = &do_glue_top;
		}
		else if (strcmp(design_name, "aes") == 0)
		{
			event_function = &do_next_aes_event_block_input;
			header_function = &do_aes_top;
		}
		else if (strcmp(design_name, "lziv") == 0)
		{
			event_function = &do_next_lziv_event_block_input;
			header_function = NULL;
		}
		else if (strcmp(design_name, "bridge") == 0)
		{
			event_function = &do_next_bridge_event_input;
			header_function = NULL;
		}
		else if (strcmp(design_name, "2dconv") == 0)
		{
			event_function = &do_next_2dconv_event_block_input;
			header_function = NULL;
		}
	
		/* add another workload to the global set of input workloads */
		initialize_workload(current_workload);
	
		/* copy in the workload details */
		current_workload->design_name = design_name;
		current_workload->workload_number = workload_number;
	
		/* put the default reset header */
		add_reset_header_to_workload(current_workload);
	
		if (header_function != NULL)
		{
			/* IF - the workload has a header function that does a preamble then execute it */
			(*header_function)(current_workload, default_workloads.workload_waves[i]);
		}
	
		/* create the pulse wave and the associated events */
		pulse_wave_of_events(	default_workloads.workload_waves[index_found]->X_ns, 
					default_workloads.workload_waves[index_found]->X_deviation, 
					default_workloads.workload_waves[index_found]->Y_ns,
				       	default_workloads.workload_waves[index_found]->Y_deviation, 
					default_workloads.workload_waves[index_found]->Z_ns,
				       	default_workloads.workload_waves[index_found]->Z_deviation, 
					default_workloads.workload_waves[index_found]->total_time,
					event_function,
					current_workload, 
					default_workloads.workload_waves[i]);

		return current_workload;
	}

	return NULL;
}

/*---------------------------------------------------------------------------------------------
 * (function: pulse_wave_of_events)
 * Data bursts happen every X time_units
 * Each data burst has data generated for Y time_units
 * For a data burst the data is generated every Z time_units
 *
 * So X and Y define something like a step function, and Z defines the frequency of data when
 * the burst is happening
 *
 * Diagram of the pulse
 * ________-------________-------________-------________-------
 * .     X       .
 *         .  Y  .
 *
 * This randomness is created by the passed in parameters and then for each benchmark the 
 * events are generated.
 *-------------------------------------------------------------------------------------------*/
void pulse_wave_of_events(
		mpz_t* X, int X_percent_deviation, 
		mpz_t* Y, int Y_percent_deviation,
		mpz_t* Z, int Z_percent_deviation,
		mpz_t *total_time,
		void (*function_ptr)(mpz_t*, workload_t*, workload_waves_t*, short),
		workload_t* current_workload,
		workload_waves_t *wave)
{
	mpz_t current_time;
	mpz_t burst_time;
	mpz_t next_event_time;
	int random_number1;
	int random_number2;
	short first_time = TRUE;

	assert(X_percent_deviation < 100);
	assert(Y_percent_deviation < 100);
	assert(Z_percent_deviation < 100);

	mpz_init (burst_time);
	mpz_init (next_event_time);
	
	mpz_init_set_str(current_time, "10000000", 10); // start at 10 ms into timeline to allow for resets

	while (mpz_cmp(*total_time, current_time) > 0)
	{
		/* start a burst */
		random_number1 = my_rand()%Y_percent_deviation+1;

		/* get the time of the burst */
		mpz_mul_ui(burst_time, *Y, 100-random_number1); // I shrink the time by a simple 80/100 = new_time/max_time
		mpz_cdiv_q_ui(burst_time, burst_time, 100);	
		mpz_add(burst_time, burst_time, current_time);

		/* keep on generating events until the burst time is up */
		while(mpz_cmp(burst_time, current_time) > 0)
		{
			random_number2 = my_rand()%Z_percent_deviation+1;

			/* get the variation on the y time */
			mpz_mul_ui(next_event_time, *Z, 100+random_number2);	
			mpz_cdiv_q_ui(next_event_time, next_event_time, 100);	
			mpz_add(current_time, next_event_time, current_time);

			/* do the possible events by calling the passed in event generation function */
			(*function_ptr)(&current_time, current_workload, wave, first_time);

			/* first time is to initialize the called function properly */
			if (first_time == TRUE)
			{
				first_time = FALSE;
			}
		}

		/* add the time off period ... this noting that Ytime+timeoff = Xtime */
		/* get the Y time */
		mpz_mul_ui(burst_time, *Y, 100-random_number1);	
		mpz_cdiv_q_ui(burst_time, burst_time, 100);	

		/* subtract Y time from X time */
		mpz_sub(burst_time, *X, burst_time);

		/* change X time by some variation */
		random_number2 = my_rand()%X_percent_deviation+1;
		mpz_mul_ui(burst_time, burst_time, 100-random_number2);	
		mpz_cdiv_q_ui(burst_time, burst_time, 100);	

		/* add the X time to the current time */
		mpz_add(current_time, current_time, burst_time);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: add_reset_header_to_workload)
 * adds the reset signal and the reset on to initialize the chip
 *-------------------------------------------------------------------------------------------*/
void add_reset_header_to_workload(workload_t *this_workload)
{
	int i;
	char *temp_string;
	int num_signals = 4; 
	char *resource_name;
	char **value;

	mpz_t *time;

	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "0", 10); // 0 us
	
	/* set the reset value */
	value = (char**)malloc(sizeof(char*)*1);
	value[0] = (char*)malloc(sizeof(char)*2);

	sprintf(value[0], "1");

	resource_name = (char*)malloc(sizeof(char)*6);
	sprintf(resource_name, "reset");
	add_event_to_event_list(this_workload, time, -1, resource_name, value, 1, INPUT_EVENT, environment.minimum_operating_speed);

	/* end the reset */
	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "30000", 10); // 30 us

	value = (char**)malloc(sizeof(char*)*1);
	value[0] = (char*)malloc(sizeof(char)*2);

	sprintf(value[0], "0");

	resource_name = (char*)malloc(sizeof(char)*6);
	sprintf(resource_name, "reset");
	add_event_to_event_list(this_workload, time, -1, resource_name, value, 1, INPUT_EVENT, environment.minimum_operating_speed);
}
