/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "gmp.h"
#include "util.h"
#include "wave_generation.h"
#include "wg_glue_logic.h"
#include "fm_glue_logic.h"

// the states for the glue logic state machine
#define RESET 0
#define STATE_IDLE 4
#define WAIT 1
#define PRO_SLOW 2
#define PRO_FAST 3

// state storage
short state1;
short state2;
short state3;

/*---------------------------------------------------------------------------------------------
 * (function: do_next_glue_logic_input)
 * The glue logic event generation works on generating a bus signal that moves the 
 * three core state machines into different states.
 *-------------------------------------------------------------------------------------------*/
void do_next_glue_logic_input(mpz_t *current_time, workload_t* this_workload, workload_waves_t *wave, short first_time)
{
	int i;
	char *temp_string;
	int num_states = 3; 
	int chance_of_state_idle_change = wave->workload_value[0]; // % chance of idle signal happening, defined in wave
	char **value;
	int state_machine_changes = my_rand() % num_states;
	int state_idle = my_rand() % 100;
	
	/* a state machine is chosen.  Depending on it's state, we send the appropriate signal, which might affect other state machines.
	 * We need to do this so we don't send to many idle signals (RESET state) */
	if (state_machine_changes == 0)
	{
		/* IF - this is the first stat machine */
		if (state1 == STATE_IDLE)  
		{
			/* IF - it is in idle then we leave idle.  Any other state machines in idle will follow suit */
			/* go to wait */
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "0");
			add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			/* signal changes all states to a wait state */
			if (state1 == STATE_IDLE)
				state1 = WAIT;
			if (state2 == STATE_IDLE)
				state2 = WAIT;
			if (state3 == STATE_IDLE)
				state3 = WAIT;
		}
		else if (state1 == WAIT)
		{
			/* IF - we're in the wait state, we move to a new state...idle is passed some times and affects all state machines */
			int val_to_write = my_rand() % 3 + 1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);

			if (state_idle < chance_of_state_idle_change)
			{
				/* go to STATE_IDLE */
				/* this is to go to the idle states in all three state counters */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*2);
				sprintf(value[0], "1");
				add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
				/* signal changes all states */
				if ((state1 == WAIT) || (state1 == STATE_IDLE))
					state1 = STATE_IDLE;
				if ((state2 == WAIT) || (state2 == STATE_IDLE))
					state2 = STATE_IDLE;
				if ((state3 == WAIT) || (state3 == STATE_IDLE))
					state3 = STATE_IDLE;
			}
			else
			{
				/* move to a new state */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*9);

				/* address of this state machine */
				sprintf(value[0], "000000%s", temp_string);
				// PAJ check this isn't an error
				add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);

				state1 = val_to_write;
			}
	
			free(temp_string);
		}
		else 
		{
			/* a write to the bus for counter 1 (0001, 0010, 0011), 2, or 3 */
			/* in this case last 8 bits are write bits */
			int val_to_write = my_rand() % 3+1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);
	
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*9);
			sprintf(value[0], "000000%s", temp_string);
	
			add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			free(temp_string);

			state1 = val_to_write;
		}
	}
	else if (state_machine_changes == 1)
	{
		if (state2 == STATE_IDLE)
		{
			/* go to wait */
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "0");
			add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			/* signal changes all states */
			if (state1 == STATE_IDLE)
				state1 = WAIT;
			if (state2 == STATE_IDLE)
				state2 = WAIT;
			if (state3 == STATE_IDLE)
				state3 = WAIT;
		}
		else if (state2 == WAIT)
		{
			int val_to_write = my_rand() % 3 + 1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);

			if (state_idle < chance_of_state_idle_change)
			{
				/* go to STATE_IDLE */
				/* this is to go to the idle states in all three state counters */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*2);
				sprintf(value[0], "1");
				add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
				/* signal changes all states */
				if ((state1 == WAIT) || (state1 == STATE_IDLE))
					state1 = STATE_IDLE;
				if ((state2 == WAIT) || (state2 == STATE_IDLE))
					state2 = STATE_IDLE;
				if ((state3 == WAIT) || (state3 == STATE_IDLE))
					state3 = STATE_IDLE;
			}
			else
			{
				/* move to new or wait */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*9);

				sprintf(value[0], "000001%s", temp_string);
				add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);

				state2 = val_to_write;
			}
	
			free(temp_string);
		}
		else 
		{
			/* a write to the bus for counter 1 (0001, 0010, 0011), 2, or 3 */
			/* in this case last 8 bits are write bits */
			int val_to_write = my_rand() % 3+1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);
	
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*9);
			sprintf(value[0], "000001%s", temp_string);
	
			add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			free(temp_string);

			state2 = val_to_write;
		}
	}
	else if (state_machine_changes == 2)
	{
		if (state3 == STATE_IDLE)
		{
			/* go to wait */
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "0");
			add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			/* signal changes all states */
			if (state1 == STATE_IDLE)
				state1 = WAIT;
			if (state2 == STATE_IDLE)
				state2 = WAIT;
			if (state3 == STATE_IDLE)
				state3 = WAIT;
		}
		else if (state3 == WAIT)
		{
			int val_to_write = my_rand() % 3 + 1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);

			if (state_idle < chance_of_state_idle_change)
			{
				/* go to STATE_IDLE */
				/* this is to go to the idle states in all three state counters */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*2);
				sprintf(value[0], "1");
				add_event_to_event_list(this_workload, current_time, IDLE, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
				/* signal changes all states */
				if ((state1 == WAIT) || (state1 == STATE_IDLE))
					state1 = STATE_IDLE;
				if ((state2 == WAIT) || (state2 == STATE_IDLE))
					state2 = STATE_IDLE;
				if ((state3 == WAIT) || (state3 == STATE_IDLE))
					state3 = STATE_IDLE;
			}
			else
			{
				/* move to new or wait */
				value = (char**)malloc(sizeof(char*)*1);
				value[0] = (char*)malloc(sizeof(char)*9);

				sprintf(value[0], "000011%s", temp_string);
				
				add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);

				state3 = val_to_write;
			}
	
			free(temp_string);
		}
		else 
		{
			/* a write to the bus for counter 1 (0001, 0010, 0011), 2, or 3 */
			/* in this case last 8 bits are write bits */
			int val_to_write = my_rand() % 3+1; // max size 2 bits and no 00
			char *temp_string = convert_int_to_bit_string(val_to_write, 2);
	
			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*9);
			sprintf(value[0], "000011%s", temp_string);
	
			add_event_to_event_list(this_workload, current_time, BUS, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
	
			free(temp_string);

			state3 = val_to_write;
		}
	}
//	printf("state1 = %d, state2 = %d, state3 = %d\n", state1, state2, state3);
}

/*---------------------------------------------------------------------------------------------
 * (function: do_glue_top)
 *-------------------------------------------------------------------------------------------*/
void do_glue_top(workload_t *current_workload, workload_waves_t *wave)
{
	/* REcord that the machine starts in IDLE */
	state1 = STATE_IDLE;
	state2 = STATE_IDLE;
	state3 = STATE_IDLE;
}

