/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "types.h"
#include "gmp.h"
#include "util.h"
#include "wave_generation.h"
#include "wg_portkey.h"
#include "fm_portkey.h"

void create_portkey_header(workload_t *this_workload, const char *direction_string, const char *interrupt_string, const char *value_string, const char *control_string);

/*---------------------------------------------------------------------------------------------
 * (function: do_next_portkey_event)
 * Three current designs for the portkey expander ... GPIO, keypad 8*7, and keypad 4*4
 *
 * Depending on which of these we are generating I/O stimuli for, use random numbers
 * and percentages in the waveform parameters to generate signals.
 *-------------------------------------------------------------------------------------------*/
void do_next_portkey_event(mpz_t *current_time, workload_t *this_workload, workload_waves_t *wave, short first_time)
{
	int i,j;
	char *temp_string;
	static short eight_portkey_vals[8];
	static short keypad[56];
	static short is_pressed;

	char **value;
		
	if (first_time == TRUE)
	{
		/* initialize all the input portkey_pins from the external environment */
		for (j = 0; j < 8; j++)
		{
			eight_portkey_vals[j] = my_rand() % 2;
		}
		for (j = 0; j < 56; j++)
		{
			keypad[j] = FALSE;
		}
		is_pressed = FALSE;
	}

	if (wave->workload_value[0] == 0) // GPIO only
	{
		int event_type = my_rand()%100;

		assert(wave->num_workload_value == 2);

		if (event_type < wave->workload_value[1]) // This is the parameter read in the workload that decides the percentage change between read and write
		{
			/* IF a portkey read */
			/* in this case first 8 bits are read bits */
			int pin_num = my_rand()%8;
			/* toggle value so it's different then before and will cause an interrupt */
			eight_portkey_vals[pin_num] = (eight_portkey_vals[pin_num] + 1) % 2;

			value = (char**)malloc(sizeof(char*)*2);
			value[0] = (char*)malloc(sizeof(char)*3);
			value[1] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "%d", pin_num);
			sprintf(value[1], "%d", eight_portkey_vals[pin_num]);

			add_event_to_event_list(this_workload, current_time, GPIO_PIN, NULL, value, 2, INPUT_EVENT, environment.minimum_operating_speed);
		}
		else if (event_type >= wave->workload_value[1])
		{
			/* ELSE IF a write to the value register */
			/* in this case last 8 bits are write bits */
			int val_to_write = my_rand();
			char *temp_string = convert_int_to_bit_string(val_to_write, 8);

			value = (char**)malloc(sizeof(char*)*1);
			value[0] = (char*)malloc(sizeof(char)*33);
			sprintf(value[0], "1110000000000000%s", temp_string);

			add_event_to_event_list(this_workload, current_time, SPI, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);

			free(temp_string);
		}
	}
	else if (wave->workload_value[0] == 1) // 8x7
	{
		if (is_pressed == FALSE)
		{
			/* IF - the key isn't pressed then generate a new press */
			int row = my_rand()%8;
			int column = my_rand()%7;
					
			assert(wave->num_workload_value == 1);
	
			value = (char**)malloc(sizeof(char*)*2);
			value[0] = (char*)malloc(sizeof(char)*2);
			value[1] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "%d", row);
			sprintf(value[1], "%d", column);
	
			/* random keyboard presses */
			add_event_to_event_list(this_workload, current_time, KEYPAD_ON, NULL, value, 2, INPUT_EVENT, environment.minimum_operating_speed);

			keypad[row+column*8] = TRUE;
			is_pressed = TRUE;
		}
		else
		{
			/* ELSE - we need to stop pressing the key */
			int row;
			int column;

			for (i = 0; i < 56; i++)
			{
				if (keypad[i] == TRUE)
				{
					row = i%8;
					column = i/8;
					keypad[i] = FALSE;
					is_pressed = FALSE;
					break;
				}
			}

			value = (char**)malloc(sizeof(char*)*2);
			value[0] = (char*)malloc(sizeof(char)*2);
			value[1] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "%d", row);
			sprintf(value[1], "%d", column);
	
			/* random keyboard release */
			add_event_to_event_list(this_workload, current_time, KEYPAD_OFF, NULL, value, 2, INPUT_EVENT, environment.minimum_operating_speed);
		}
	}
	else if (wave->workload_value[0] == 2) // 4x4
	{
		if (is_pressed == FALSE)
		{
			/* IF - the key isn't pressed then generate a new press */
			int row = my_rand()%8;
			int column = my_rand()%7;
					
			assert(wave->num_workload_value == 1);
	
			value = (char**)malloc(sizeof(char*)*2);
			value[0] = (char*)malloc(sizeof(char)*2);
			value[1] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "%d", row);
			sprintf(value[1], "%d", column);
	
			/* random keyboard presses */
			add_event_to_event_list(this_workload, current_time, KEYPAD_ON, NULL, value, 2, INPUT_EVENT, environment.minimum_operating_speed);

			keypad[row+column*4] = TRUE;
			is_pressed = TRUE;
		}
		else
		{
			/* ELSE - we need to stop pressing the key */
			int row;
			int column;

			for (i = 0; i < 56; i++)
			{
				if (keypad[i] == TRUE)
				{
					row = i%4;
					column = i/4;
					keypad[i] = FALSE;
					is_pressed = FALSE;
					break;
				}
			}

			value = (char**)malloc(sizeof(char*)*2);
			value[0] = (char*)malloc(sizeof(char)*2);
			value[1] = (char*)malloc(sizeof(char)*2);
			sprintf(value[0], "%d", row);
			sprintf(value[1], "%d", column);
	
			/* random keyboard release */
			add_event_to_event_list(this_workload, current_time, KEYPAD_OFF, NULL, value, 2, INPUT_EVENT, environment.minimum_operating_speed);
		}
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: do_portkey_top)
 *-------------------------------------------------------------------------------------------*/
void do_portkey_top(workload_t *current_workload, workload_waves_t *wave)
{
	mpz_t *time;
	char **value;

	/* add some conditions to this benchmark */
	current_workload->conditions->num_if_resource_pairs = 0;
	current_workload->conditions->then_resources = NULL;

	current_workload->conditions->num_if_resource_pairs = 2;

	/* allocate another spot on the ifs of resources */
	current_workload->conditions->ifs = (resource_t **)malloc(sizeof(resource_t*)*2);
	current_workload->conditions->then_resources = (resource_t **)malloc(sizeof(resource_t*)*2);
	current_workload->conditions->then_times = (mpz_t **)malloc(sizeof(mpz_t*)*2);

	current_workload->conditions->then_times[0] = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*current_workload->conditions->then_times[0], "20000", 10); // 20 us

	/* this is interrupt0 response which states to send an spi read of the keypad */
	current_workload->conditions->ifs[0] = (resource_t*)malloc(sizeof(resource_t));
	current_workload->conditions->ifs[0]->name = (char*)malloc(sizeof(char)*11);
	sprintf(current_workload->conditions->ifs[0]->name, "interrupt0");
	current_workload->conditions->ifs[0]->value = NULL;
	current_workload->conditions->ifs[0]->num_values = 0;

	current_workload->conditions->then_resources[0] = (resource_t*)malloc(sizeof(resource_t));
	current_workload->conditions->then_resources[0]->name = (char*)malloc(sizeof(char)*4);
	sprintf(current_workload->conditions->then_resources[0]->name, "spi");
	current_workload->conditions->then_resources[0]->value = (char**)malloc(sizeof(char*)*1);
	current_workload->conditions->then_resources[0]->value[0] = (char*)malloc(sizeof(char)*9);
	sprintf(current_workload->conditions->then_resources[0]->value[0], "00010000");
	current_workload->conditions->then_resources[0]->num_values = 1;

	current_workload->conditions->then_times[1] = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*current_workload->conditions->then_times[1], "20000", 10); // 20 us

	/* this is interrupt1 response which states to send an spi read of the value register */
	current_workload->conditions->ifs[1] = (resource_t*)malloc(sizeof(resource_t));
	current_workload->conditions->ifs[1]->name = (char*)malloc(sizeof(char)*11);
	sprintf(current_workload->conditions->ifs[1]->name, "interrupt1");
	current_workload->conditions->ifs[1]->value = NULL;
	current_workload->conditions->ifs[1]->num_values = 0;

	current_workload->conditions->then_resources[1] = (resource_t*)malloc(sizeof(resource_t));
	current_workload->conditions->then_resources[1]->name = (char*)malloc(sizeof(char)*4);
	sprintf(current_workload->conditions->then_resources[1]->name, "spi");
	current_workload->conditions->then_resources[1]->value = (char**)malloc(sizeof(char*)*1);
	current_workload->conditions->then_resources[1]->value[0] = (char*)malloc(sizeof(char)*9);
	sprintf(current_workload->conditions->then_resources[1]->value[0], "01100000");
	current_workload->conditions->then_resources[1]->num_values = 1;

	/* now setup the registers properly for the mode we are in */
	if (wave->workload_value[0] == 0) // GPIO
	{
		/* last 8 bits are outputs
		 * alternating first input bits are interrupted
		 * value is alternating for outputs
		 * no keypad */
		create_portkey_header(current_workload, "101000000000000011111111", "110000000101010100000000", "111000000000000010101010", "100000000000000000000000");
	}
	else if (wave->workload_value[0] == 1) // 8x7 keypad
	{
		/* no portkeys
		 * no interrupts
		 * no value
		 * 8x7 keypad */
		create_portkey_header(current_workload, "101000000000000000000000", "110000000000000000000000", "111000000000000000000000", "100000000001111100000000");
	}
	else if (wave->workload_value[0] == 2) // 4x4 keypad
	{
		/* no portkeys
		 * no interrupts
		 * no value
		 * 4x4 keypad */
		create_portkey_header(current_workload, "101000000000000000000000", "110000000000000000000000", "111000000000000000000000", "100000000010001100000000");
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: create_portkey_header)
 * Initializes the SPI writes to the registers to setup the device
 *-------------------------------------------------------------------------------------------*/
void create_portkey_header(workload_t *this_workload, const char *direction_string, const char *interrupt_string, const char *value_string, const char *control_string)
{
	char **value;
	mpz_t *time;

	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "1000000", 10); // 1000 us
		
	value = (char**)malloc(sizeof(char*)*2);
	value[0] = (char*)malloc(sizeof(char)*(strlen(direction_string)+1));
	sprintf(value[0], "%s", direction_string);

	add_event_to_event_list(this_workload, time, SPI, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);


	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "3000000", 10); // 3000 us
		
	value = (char**)malloc(sizeof(char*)*2);
	value[0] = (char*)malloc(sizeof(char)*(strlen(interrupt_string)+1));
	sprintf(value[0], "%s", interrupt_string);

	add_event_to_event_list(this_workload, time, SPI, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);

	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "5000000", 10); // 5000 us
		
	value = (char**)malloc(sizeof(char*)*2);
	value[0] = (char*)malloc(sizeof(char)*(strlen(value_string)+1));
	sprintf(value[0], "%s", value_string);

	add_event_to_event_list(this_workload, time, SPI, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);


	time = (mpz_t*)malloc(sizeof(mpz_t));
	mpz_init_set_str(*time, "7000000", 10); // 7000 us
		
	value = (char**)malloc(sizeof(char*)*2);
	value[0] = (char*)malloc(sizeof(char)*(strlen(control_string)+1));
	sprintf(value[0], "%s", control_string);

	add_event_to_event_list(this_workload, time, SPI, NULL, value, 1, INPUT_EVENT, environment.minimum_arrival_rate_on_serial_interfaces);
}
