/* 
 * Copyright (c) 2008 - Peter Jamieson
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 * */
#include <stdlib.h>
#include "globals.h"
#include "types.h"
#include "util.h"
#include "gmp.h"
#include "workload_utils.h"

/*---------------------------------------------------------------------------------------------
 * (function: initialize_workload)
 *-------------------------------------------------------------------------------------------*/
void initialize_workload(workload_t *workload)
{
	workload->events = (events_t*)malloc(sizeof(events_t));
	workload->events->events = NULL;
	workload->events->num_events = 0;
	workload->design_name = NULL;
	workload->workload_number = -1;
	workload->conditions = (condition_t*)malloc(sizeof(condition_t));
	workload->conditions->num_if_resource_pairs = 0;
	workload->conditions->ifs = NULL;
	workload->conditions->then_resources = NULL;
	workload->conditions->then_times = NULL;
}

/*---------------------------------------------------------------------------------------------
 * (function: free_workload)
 *-------------------------------------------------------------------------------------------*/
void free_event(event_t *event);
void free_resource(resource_t *resource);
void free_workload(workload_t *workload)
{
	int i;
	for (i = 0; i < workload->events->num_events; i++)
	{
		if (workload->events->events[i] != NULL)
		{
			if (workload->events->events[i]->time != NULL)
			{
				mpz_clear(*workload->events->events[i]->time);
				free(workload->events->events[i]->time);
				workload->events->events[i]->time = NULL;
			}

			free_event(workload->events->events[i]);
			free(workload->events->events[i]);
			workload->events->events[i] = NULL;
		}
	}
	if (workload->events->events != NULL)
	{
		free(workload->events->events);	
		workload->events->events = NULL;
	}

	for (i = 0; i < workload->conditions->num_if_resource_pairs; i++)
	{
		free_resource(workload->conditions->ifs[i]);
		free(workload->conditions->ifs[i]);
		free_resource(workload->conditions->then_resources[i]);
		free(workload->conditions->then_resources[i]);
		mpz_clear(*workload->conditions->then_times[i]);
		free(workload->conditions->then_times[i]);
	}
	if (workload->conditions->ifs != NULL)
		free(workload->conditions->ifs);
	if (workload->conditions->then_resources != NULL)
		free(workload->conditions->then_resources);
	if (workload->conditions->then_times != NULL)
		free(workload->conditions->then_times);
}
void free_event(event_t *event)
{
	int i;
	for (i = 0; i < event->num_resources; i++)
	{
		free_resource(event->resources[i]);
		free(event->resources[i]);
	}
}
void free_resource(resource_t *resource)
{
	int i;
	for (i = 0; i < resource->num_values; i++)
	{
		if (resource->value[i] != NULL)
		{
			free(resource->value[i]);
		}
	}
	if (resource->value != NULL)
	{
		free(resource->value);
	}
	if (resource->name != NULL)
	{
		free(resource->name);
	}
}

/*---------------------------------------------------------------------------------------------
 * (function: add_event_to_event_list)
 * adds a one resource event to a workload.
 *-------------------------------------------------------------------------------------------*/
void add_event_to_event_list(workload_t *to_add_to, mpz_t *time, short resource_name_id, char *resource_name, char **values, int num_values, short event_type, mpz_t *clock)
{
	int i;
	mpz_t temp;

	/* adjust based on the clock so we roughly allign with a clock domain
	 * (Clock - (Current Time % Clock)) + Current Time
	 */
	mpz_init(temp);
	mpz_fdiv_r(temp, *time, *clock);
	if (mpz_cmp_ui(temp, 0) != 0)
	{
		mpz_sub(temp, *clock, temp);
		mpz_add(*time, *time, temp);
	}
	mpz_clear(temp);

	if ((to_add_to->events->num_events == 0) || ((to_add_to->events->num_events > 0) && (mpz_cmp(*(to_add_to->events->events[to_add_to->events->num_events-1]->time), *(time)) != 0)))
	{
		/* IF - this event doesn't have the same time as the last event then add a new event */

		to_add_to->events->num_events++;
		to_add_to->events->events = (event_t**)realloc(to_add_to->events->events, sizeof(event_t*)*to_add_to->events->num_events);
		to_add_to->events->events[to_add_to->events->num_events-1] = (event_t*)malloc(sizeof(event_t));
		to_add_to->events->events[to_add_to->events->num_events-1]->num_resources = 0;
		to_add_to->events->events[to_add_to->events->num_events-1]->resources = NULL;

		to_add_to->events->events[to_add_to->events->num_events-1]->time = (mpz_t *)malloc(sizeof(mpz_t));
		mpz_init(*(to_add_to->events->events[to_add_to->events->num_events-1]->time));
		mpz_set(*(to_add_to->events->events[to_add_to->events->num_events-1]->time), *time);

		/* store the event type */
		to_add_to->events->events[to_add_to->events->num_events-1]->type = event_type;
	}

	/* add a new resource */
	to_add_to->events->events[to_add_to->events->num_events-1]->num_resources++;
	to_add_to->events->events[to_add_to->events->num_events-1]->resources = (resource_t **)realloc( to_add_to->events->events[to_add_to->events->num_events-1]->resources, sizeof(resource_t*)*to_add_to->events->events[to_add_to->events->num_events-1]->num_resources);

	/* allocate next record */
	to_add_to->events->events[to_add_to->events->num_events-1]->resources[to_add_to->events->events[to_add_to->events->num_events-1]->num_resources-1] = (resource_t *)malloc(sizeof(resource_t)); 

	/* store the name */
	to_add_to->events->events[to_add_to->events->num_events-1]->resources[to_add_to->events->events[to_add_to->events->num_events-1]->num_resources-1]->default_name = resource_name_id; 
	/* store the name */
	to_add_to->events->events[to_add_to->events->num_events-1]->resources[to_add_to->events->events[to_add_to->events->num_events-1]->num_resources-1]->name = resource_name; 

	/* store the values */
	to_add_to->events->events[to_add_to->events->num_events-1]->resources[to_add_to->events->events[to_add_to->events->num_events-1]->num_resources-1]->num_values = num_values; 
	to_add_to->events->events[to_add_to->events->num_events-1]->resources[to_add_to->events->events[to_add_to->events->num_events-1]->num_resources-1]->value = values;
}

/*---------------------------------------------------------------------------------------------
 * (function: generate_the_output_workload_file)
 * Given a workload sturcture this will create the output XML file
 *-------------------------------------------------------------------------------------------*/
void generate_an_output_workload_file(char *file_name, workload_t *output_workload, char *design_name, int workload_number)
{
	FILE *file_to_write;
	int i, j, k;
	char *temp_string;
	workload_waves_t *wave;

	int index_found = -1;

	/* find the index of waves with the same name as the design name */
	for (i = 0; i < default_workloads.num_workload_waves; i++)
	{
		if ((strcmp(default_workloads.workload_waves[i]->design_name, design_name) == 0) && (atoi(default_workloads.workload_waves[i]->name) == workload_number))
		{
			index_found = i;
			wave = default_workloads.workload_waves[i];
			break;

		}
	}

	file_to_write = fopen(file_name, "w");
	if (file_to_write == NULL)
	{
		fprintf(stderr, "Can't open filename: %s\n", file_name);
	}

	/* output the header to the xml file which has some details about the environment, the design, and the waveform that generated it */
	fprintf(file_to_write, "<workload>\n");
	fprintf(file_to_write, "\t<design_name>%s</design_name>\n", design_name);
	fprintf(file_to_write, "\t<workload_number>%d</workload_number>\n", workload_number);
	temp_string = mpz_get_str(NULL, 10, *(default_workloads.workload_waves[index_found]->Z_ns));
	fprintf(file_to_write, "\t<minimum_arrival_rate>%sns</minimum_arrival_rate>\n", temp_string);
	free(temp_string);
	temp_string = mpz_get_str(NULL, 10, *(environment.minimum_operating_speed));
	fprintf(file_to_write, "\t<minimum_operating_speed>%sns</minimum_operating_speed> <!-- can be related to clock speed of system-->\n", temp_string);
	free(temp_string);
	temp_string = mpz_get_str(NULL, 10, *(environment.minimum_sampling_speed));
	fprintf(file_to_write, "\t<minimum_sampling_speed>%sns</minimum_sampling_speed> <!-- heart beat of the system -->\n", temp_string);
	free(temp_string);
	temp_string = mpz_get_str(NULL, 10, *(environment.minimum_arrival_rate_on_serial_interfaces));
	fprintf(file_to_write, "\t<minimum_arrival_rate_on_serial_interfaces>%sns</minimum_arrival_rate_on_serial_interfaces> <!-- can realte to the serial clock -->\n", temp_string);
	free(temp_string);
	temp_string = mpz_get_str(NULL, 10, *(environment.minimum_arrival_rate_on_parallel_interfaces));
	fprintf(file_to_write, "\t<minimum_arrival_rate_on_parallel_interfaces>%sns</minimum_arrival_rate_on_parallel_interfaces> <!-- can relate to the parallel clock -->\n", temp_string);
	free(temp_string);

	if (mpz_cmp(*(default_workloads.workload_waves[index_found]->Z_ns), *(environment.minimum_operating_speed)) < 0)
	{
		/* if the Z arrival rate is less than the max clock speed, this will likely fail */
		fprintf(file_to_write, "\t<!-- WARNING!!! For the given clock speed and the Z min arrival rate, this workload will fail to work -->\n");
	}

	/* print out the appropriate specs for the design if it has different modes */
	if (strcmp(design_name, "portkey") == 0)
	{
		if (wave->workload_value[0] == 0) // portkey only
		{
			fprintf(file_to_write, "\t<additional_details>16gpios, pin0: read_int, pin1: read, pin2: read_int, pin3: read, pin4: read_int, pin5: read, pin6: read_int, pin7: read, pin8: write, pin9: write, pin10: write, pin11: write, pin12: write, pin13: write, pin14: write, pin15: write</additional_details>\n");
		}
		else if (wave->workload_value[0] == 1) // 8x7
		{
			fprintf(file_to_write, "\t<additional_details>keypad: 8x7</additional_details>\n");
		}
		else if (wave->workload_value[0] == 2) // 4x4
		{
			fprintf(file_to_write, "\t<additional_details>keypad: 4x4</additional_details>\n");
		}
	}

	if (output_workload->conditions->num_if_resource_pairs > 0)
	{
		fprintf(file_to_write, "\t<conditions>\n");
		/* now go through all the events and print out there value */
		for (i = 0; i < output_workload->conditions->num_if_resource_pairs; i++)
		{
			fprintf(file_to_write, "\t\t<condition>\n");
			temp_string = mpz_get_str(NULL, 10, *(output_workload->conditions->then_times[i]));
			fprintf(file_to_write, "\t\t\t<time>%sns</time>\n", temp_string);
			free(temp_string);
	
			fprintf(file_to_write, "\t\t\t<if>\n\t\t\t\t<resource>\n\t\t\t\t\t<name>%s</name>\n", output_workload->conditions->ifs[i]->name);
	
			for (j = 0; j < output_workload->conditions->ifs[i]->num_values; j++)
			{
				fprintf(file_to_write, "\t\t\t\t\t<value%d>%s</value%d>\n", j, output_workload->conditions->ifs[i]->value[j], j);
			}
			fprintf(file_to_write, "\t\t\t\t</resource>\n");
			fprintf(file_to_write, "\t\t\t</if>\n");
	
			fprintf(file_to_write, "\t\t\t<resource>\n\t\t\t\t<name>%s</name>\n", output_workload->conditions->then_resources[i]->name);
			for (j = 0; j < output_workload->conditions->then_resources[i]->num_values; j++)
			{
				fprintf(file_to_write, "\t\t\t\t<value%d>%s</value%d>\n", j, output_workload->conditions->then_resources[i]->value[j], j);
			}
			fprintf(file_to_write, "\t\t\t</resource>\n");
	
			fprintf(file_to_write, "\t\t</condition>\n");
		}
		fprintf(file_to_write, "\t</conditions>\n");
	}

	fprintf(file_to_write, "\t<events>\n");
	/* now go through all the events and print out there value */
	for (i = 0; i < output_workload->events->num_events; i++)
	{
		fprintf(file_to_write, "\t\t<event>\n");
		temp_string = mpz_get_str(NULL, 10, *(output_workload->events->events[i]->time));

		fprintf(file_to_write, "\t\t\t<time>%sns</time>\n", temp_string);
		for (j = 0; j < output_workload->events->events[i]->num_resources; j++)
		{
			fprintf(file_to_write, "\t\t\t<resource>\n");
			if (output_workload->events->events[i]->resources[j]->name != NULL)
			{
				fprintf(file_to_write, "\t\t\t\t<name>%s</name>\n", output_workload->events->events[i]->resources[j]->name);
			}
			else 
			{
				char *temp_string;
				assert(output_workload->events->events[i]->resources[j]->default_name != -1);

				temp_string = get_default_resource_name(output_workload->events->events[i]->resources[j]->default_name);
				fprintf(file_to_write, "\t\t\t\t<name>%s</name>\n", temp_string);
				free(temp_string);
			}

			for (k = 0; k < output_workload->events->events[i]->resources[j]->num_values; k++)
			{
				fprintf(file_to_write, "\t\t\t\t<value%d>%s</value%d>\n", k, output_workload->events->events[i]->resources[j]->value[k], k);
			}

			fprintf(file_to_write, "\t\t\t</resource>\n");
		}
		fprintf(file_to_write, "\t\t</event>\n");
		free(temp_string);
	}

	/* close the xml file */
	fprintf(file_to_write, "\t</events>\n");
	fprintf(file_to_write, "</workload>\n");

	fclose(file_to_write);
}

/*---------------------------------------------------------------------------------------------
 * (function: get_default_resource_name)
 * This function is used for repetitive resource names.  This allows us to store a constant
 * instead of a full string.  This is where the string is recovered.
 *
 * Note: you need to free the allocated string.
 *-------------------------------------------------------------------------------------------*/
char *get_default_resource_name(short default_name)
{
	char *return_string = (char*)malloc(sizeof(char)*30);

	if (default_name == SPI)
	{
		sprintf(return_string, "spi");
	}
	else if (default_name == GPIO_PIN)
	{
		sprintf(return_string, "gpio_pin");
	}
	else if (default_name == KEYPAD_ON)
	{
		sprintf(return_string, "keypad_press");
	}
	else if (default_name == KEYPAD_OFF)
	{
		sprintf(return_string, "keypad_release");
	}
	else if (default_name == INPUT)
	{
		sprintf(return_string, "input");
	}
	else if (default_name == INPUT_BLOCK)
	{
		sprintf(return_string, "input_block");
	}
	else if (default_name == INPUT_DATA_VALID)
	{
		sprintf(return_string, "input_data_valid");
	}
	else if (default_name == BUS)
	{
		sprintf(return_string, "bus");
	}
	else if (default_name == IDLE)
	{
		sprintf(return_string, "idle");
	}
	else 
		assert(FALSE);

	return return_string;
}
